package ORG.ebiss.sara.client;

import java.io.*;
import java.net.*;
import java.awt.*;
import ORG.ebiss.sara.server.UIObject;
import ORG.ebiss.sara.server.SimValueAdapter;
import ORG.ebiss.sara.server.RemoteAgentObserver;
import ORG.ebiss.sara.server.Interaction;
import ORG.ebiss.sara.server.AgentAdapter;
import ORG.ebiss.sara.server.SimValueAdapter;
import ORG.ebiss.sara.server.Statement;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

import java.awt.event.*;
import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import com.sun.java.swing.border.*;

/**
 * Client.
 *   Player's Interface.
 *   There's a single JApplet to interface with the player. 
 * 
 */


public class Client extends JApplet implements RemoteAgentObserver {
  static final Font FONT = new Font("SansSerif", Font.PLAIN, 16);
  
  private final int WIDTH = 600; 
  private final int HEIGHT = 600;

  private final Color _color = Color.gray;
  
  private AgentAdapter adapter; // the interface to server
  private String agentName;
  private String serverName;
  private String tearOff;
  private boolean isTearOff;
  
  /* to Keep inform */
  private Hashtable informHash;
  private InformHashFrame informList;
  private Vector informs;
  
  /* to Keep interact */
  private Hashtable interactHash;
  private InteractHashFrame interactList;
  private Vector interacts;
  
  /* Multiple Situation */
  private Hashtable situationHash;
  private Vector situations;
  
  /* panels */
  private PropPanel proPanel;		/* property panel */
  private JTabbedPane sitTab;           /* situations' Tab */
  
  private JPanel panel;               /* auxiliary panel */
  private JLabel lab;                 /* auxiliary label */
  private JScrollPane scroll;         /* auxiliary scroll */
  
  private ActionPanel actPanel;
  
  protected Container pane;
  
  private int frameLocationXpoint, frameLocationYpoint;
  

    public void init() {
	agentName = getParameter("agentName");
	try {     // server interface
	    adapter = new AgentAdapter(getCodeBase().getHost(), agentName);
	} catch (UnknownHostException e) {
	    Log.error("UnknownHostException: " + e.getMessage());
	} catch (IOException e) {
	    Log.error("IOException at mkAdapter: " + e.getMessage());
	} 

	/* to keep Information */
	informHash = new Hashtable();
	informs = new Vector();

	/* to keep Interaction */
	interactHash = new Hashtable();
	interacts = new Vector();

	/* Situations */
	situationHash = new Hashtable();
	situations = new Vector();

	tearOff = getParameter("tearOff");
	isTearOff = false;
	if(tearOff != null && tearOff.equals("true"))
	  isTearOff = true;
	makeInternalPanels();		     // panels

	/* Register me to server */
	// adapter.registerAgentObserver(this);
	adapter.addObserver(this);

	frameLocationXpoint = 50;
	frameLocationYpoint = 50; 

	if(!isTearOff)
	  setSize(WIDTH, HEIGHT);
	else
	  setSize(350, 400);
	
	setBackground(_color);
    }
  
    private void makeInternalPanels(){  // make the panels
	pane = getContentPane();    

	pane.setLayout(new GridLayout(0, 1));

	JPanel northPanel = new JPanel();
	northPanel.setLayout(new BorderLayout());
	pane.add(northPanel);


	/* ToolBar */
	northPanel.add(new ToolBar(this), BorderLayout.NORTH);

	/* components other than ToolBar */

	setFont(Client.FONT);

	/* Property Panel */
	panel = new JPanel();
	panel.setFont(Client.FONT);
	panel.setBorder(new TitledBorder("Agent Properties"));
	proPanel = new PropPanel();	
	scroll =
	    new JScrollPane(proPanel,
			    ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
			    ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	panel.setLayout(new GridLayout(0,1));
	panel.add(scroll);

	northPanel.add(panel, BorderLayout.CENTER);

	/* Situation Panel */
	sitTab = new JTabbedPane();
	sitTab.setFont(Client.FONT);

	if(isTearOff){
	  /* Tear Off Situations Tab */
	  JFrame situationFrame = new JFrame(agentName + "\'s Situations");
	  situationFrame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
	  situationFrame.getContentPane().add(sitTab, BorderLayout.CENTER);
	  situationFrame.setSize(500, 500);
	  Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
	  int x = (d.width - 500)/2;
	  int y = (d.height - 500)/2;
	  situationFrame.setLocation(x, y);
	  situationFrame.setVisible(true);
	  situationFrame.validate();
	  sitTab.repaint();

	  /* Action Panel */
	  actPanel = new ActionPanel(adapter, this);
	  scroll =
	    new JScrollPane(actPanel,
			    ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
			    ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	  scroll.setBorder(new TitledBorder("Agent Actions"));
	  scroll.setBackground(Color.lightGray);
	  pane.add(scroll);

	}else{
	  /* Action Panel */
	  actPanel = new ActionPanel(adapter, this);
	  scroll =
	    new JScrollPane(actPanel,
			    ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
			    ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	  sitTab.addTab("Actions", scroll);
	  
	  panel = new JPanel();
	  panel.setLayout(new GridLayout(0,1));
	  panel.add(sitTab);
	  
	  pane.add(panel);
	}
	
	validate();
    }


    /* To Keep Information */
    public void keepInform(String name, InformFrame info){
	informHash.put(name, info);
	informs.addElement(name);
    }
    public InformFrame getInform(String s){
	InformFrame info = (InformFrame)informHash.get(s);
	return info;
    }
    public void makeInformList(){
	informList = new InformHashFrame(this, informs);    
    }

    /* To Keep Interaction */
    public void keepInteractFrame(String name,
				  InteractFrame interactF){
	interactHash.put(name, interactF);
	interacts.addElement(name);
    }
    public InteractFrame getInteract(String s){
	InteractFrame frame = (InteractFrame)interactHash.get(s);
	return frame;
    }
    public void makeInteractList(){
	interactList = new InteractHashFrame(this, interacts);    
    }
    public InteractHashFrame getInteractList(){
	return interactList;
    }
    public void disposeInteractList(){
	interactList.setVisible(false);
	interactList.dispose();
    }

    public void addSituation(SimValueAdapter svadapter,
			     String name){
	SituationPanel s = new SituationPanel(name, svadapter, this);
	situationHash.put(name, s);
	situations.addElement(name);
	sitTab.addTab(name, s);
	sitTab.validate();
    }
    public SituationPanel getSituation(String name){
	SituationPanel s = (SituationPanel)situationHash.get(name);
	return s;
    }
    public void removeSituation(SimValueAdapter svadapter, String name){
	int count = 0; 
	for (Enumeration e = situations.elements(); e.hasMoreElements();){
	    if(e.nextElement().equals(name)){
		SituationPanel s = (SituationPanel)situationHash.get(name);
		s.dispose(svadapter);
		situationHash.remove(name);
		situations.removeElementAt(count);
		if(isTearOff)
		  sitTab.removeTabAt(count);
		else
		  sitTab.removeTabAt(count+1);           // As Action Tab exists
		return;
	    }
	    else{
		count++;
	    }
	}
	sitTab.validate();
    }

  private int counter = 0;;

  public Point getFrameLocationToBe(int frameWidth, int frameHeight){
    int tmpYpoint;
    Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
    if(frameLocationXpoint >= (screen.width - frameWidth))
      frameLocationXpoint = 50;
    
    if(frameLocationYpoint <= (screen.height - frameHeight)){
      tmpYpoint = frameLocationYpoint;
    }
    else{
      counter ++;
      frameLocationYpoint = 50;
      tmpYpoint = frameLocationYpoint;
      if(counter < 3)
        frameLocationXpoint = frameLocationXpoint + 150;
      else{
        frameLocationXpoint = 50;
        counter = 0;
      }
    }
    frameLocationYpoint = frameLocationYpoint + frameHeight;
    return new Point(frameLocationXpoint, tmpYpoint);
  }
  


    /******************* RemoteAgentObserver  ******************/

    public synchronized void acceptProperty(SimValueAdapter adapter, 
					    String propertyName, 
					    UIObject prop) {
        if (propertyName.equals("EVENT")) {
	  System.out.println("Event: " + prop.toString());
        } else if (propertyName.equals("STATEMENT")) {
	  System.out.println("Statement: From: "
			     + ((Statement)prop).getSender() +
			     ", " + ((Statement)prop).getContents());
	} else {
	  proPanel.addProperty(propertyName, prop);
	  repaint();
	}
    }

    public synchronized void retractProperty(SimValueAdapter adapter,
					     String propertyName){
	proPanel.delProperty(propertyName);
	repaint();
    }
  
    public void show(AgentAdapter adapter, UIObject object) {
	System.out.println("show " + object);
    }
  
    /* handled in SituationPanel */
    public synchronized void acceptInteraction(SimValueAdapter adapter,
					       Interaction i) {
	switch(i.tag()) {
	case Interaction.TAG_DIRECT_INTERACTION: 
	    new InteractThread(adapter, i, this);
	    break;
	default: 
	    // TAG_CHAT, TAG_PROPOSE, TAG_DELEGATE, TAG_RESPONCE_TO_PROPOSAL
	    actPanel.addAction(i);
	    break;
	} 
	repaint();
    }
  
    /* handled in SituationPanel */
    public synchronized void retractInteraction(SimValueAdapter adapter,
						Interaction i) {
	  actPanel.delAction(i);
	  repaint();
    }
  
    public synchronized void acceptInformation(SimValueAdapter adapter,
					       UIObject i){
	if(informList != null){
	    informList.setVisible(false);
	    informList.dispose();
	}
	new InformThread(this, i);
    }

    public synchronized
    void acceptExtendedInformation(SimValueAdapter adapter,
				   String[] type, UIObject info) {
      System.out.println("accept extended information: type=" + type[0]);
      acceptProperty(adapter, type[0], info);
    }

    public synchronized void acceptEvent(AgentAdapter adapter, 
					 UIObject event) {
	Log.warning("Not implemented. Show situation properties instead");
    }
  
  
    public synchronized void acceptAuxiliaryObject(AgentAdapter adapter,
						   SimValueAdapter auxAdapter,
						   String[] type) {
	addSituation(auxAdapter, type[0]);
	System.out.println("an auxiliary object is created. type=" + type[0]);
    }

    public synchronized void retractAuxiliaryObject(AgentAdapter adapter,
						    SimValueAdapter auxAdapter,
						    String[] type) {
	removeSituation(auxAdapter, type[0]);
	System.out.println("an auxiliary object is destroyed. type=" + type[0]);
    }

    public String getAgentName() {
	return agentName;
    }

    /* Draw a box around this panel. 
     * Insets puts space between the panel and the box . 
     */

    public Insets getInsets() {
	return new Insets(15, 15, 15, 15);
    }
}


