package ORG.ebiss.sara.client;

import java.awt.*;
import ORG.ebiss.sara.server.UIObject;
import ORG.ebiss.sara.server.FloatParam;

import java.awt.event.*;
import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import com.sun.java.swing.border.*;

/**
 * Float parameter.
 */

class FloatInputParam extends ParamView{

  private JTextField arg;     // player's input component
  private JSlider slider;
  private float mini, max, floatValue, defaultV;
  private JButton pDotFiveB, mDotFiveB, pDotOneB, mDotOneB, resetB;

  FloatInputParam (UIObject s) {
    super(s);   // the super class constructor 

    FloatParam p = (FloatParam)s;

    if (defaultVal != null) {
      try{
	defaultV = Float.valueOf(defaultVal).floatValue();
	floatValue = defaultV;
      }catch(NumberFormatException ex){
	System.out.println(ex);
      }
      arg = new JTextField(defaultVal,15);
    } else {
      arg = new JTextField("0.0", 15);
      defaultV = (float)0.0;
      floatValue = defaultV;
    }

    mini = p.getMin();
    max = p.getMax();

    arg.addActionListener(new TextFieldInput());    
    slider =
      new JSlider(JSlider.HORIZONTAL, (int)mini, (int)max, (int)floatValue);
    slider.addChangeListener(new SlideChange());
    resetB = new JButton("RESET");
    resetB.addActionListener(new ResetValue());

    GridBagLayout gb = new GridBagLayout();
    setLayout(gb);
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;
    gbc.anchor = GridBagConstraints.NORTH;

    if (argName != null){
      argName.setFont(Client.FONT);
      gbc.gridwidth  = GridBagConstraints.REMAINDER;
      gb.setConstraints(argName, gbc);
      add(argName);
    }

    gbc.weightx = 1.0;
    gb.setConstraints(slider, gbc);
    add(slider);

    pDotFiveB = new JButton("+ " + Float.toString((float)0.5));
    mDotFiveB = new JButton("- " + Float.toString((float)0.5));
    pDotOneB = new JButton("+ " + Float.toString((float)0.1));
    mDotOneB = new JButton("- " + Float.toString((float)0.1));
    gbc.gridwidth = 1;
    gbc.weightx = 1.0;
    addButton(mDotFiveB, gb, gbc);
    addButton(mDotOneB, gb, gbc);
    addButton(pDotOneB, gb, gbc);
    gbc.gridwidth  = GridBagConstraints.REMAINDER;
    addButton(pDotFiveB, gb, gbc);

    gbc.gridwidth  = GridBagConstraints.RELATIVE;
    gb.setConstraints(arg, gbc);
    add(arg);

    JPanel panel = makeButtonPanel(resetB);

    gbc.weightx = 0.1;
    gbc.gridwidth  = GridBagConstraints.REMAINDER;
    gb.setConstraints(panel, gbc);
    add(panel);

    value = Float.toString(floatValue); 
  }

  private void addButton(JButton b, GridBagLayout gb, GridBagConstraints gbc){
    b.addActionListener(new ButtonChange());
    JPanel panel = makeButtonPanel(b);
    gb.setConstraints(panel, gbc);
    add(panel);
  }
  
  private JPanel makeButtonPanel(JButton b){
    JPanel pin = new JPanel();
    JPanel pout = new JPanel();
    pout.setLayout(new BorderLayout(0, 5));
    pin.add(b);
    pout.add(pin, BorderLayout.CENTER);
    return pout;
  }

  /* check the value */

  public boolean valid() {
    try {  // check the validy of the number
      Float.valueOf(value()); 
    } catch (java.lang.NumberFormatException ne) {
      Log.warning("You must enter a float number. Retry.");
      return false;
    } catch (NullPointerException e) {
      Log.warning("Empty argument. Retry.");
      return false;
    }
    return true;
  }

  /* Get the argument. Overhides super's method. */
  public String value() {
    value = Float.toString(floatValue); 
    return value;
  }

  public void setEnabled(boolean b){
    arg.setEditable(b);
    slider.setEnabled(b);
    pDotFiveB.setEnabled(b);
    mDotFiveB.setEnabled(b);
    pDotOneB.setEnabled(b);
    mDotOneB.setEnabled(b);
  }
    
  class TextFieldInput implements ActionListener{
    public void actionPerformed(ActionEvent e){
      try{
	floatValue = Float.valueOf(arg.getText()).floatValue();
	if(floatValue > max)
	  floatValue = max;
	if(floatValue < mini)
	  floatValue = mini;
	arg.setText(value());
	slider.setValue(new Float(floatValue).intValue());
      }catch(NumberFormatException ex){
	arg.setText(value());
	return;
      }
    }
  }
  
  class SlideChange implements ChangeListener{
    public void stateChanged(ChangeEvent e){
      float sliderValue = new Integer(slider.getValue()).floatValue();
      /*
      try{      
	floatValue = Float.valueOf(arg.getText()).floatValue();
      }catch(NumberFormatException ex){
	arg.setText(value());
	return;
      }
      */
      if(sliderValue > max)
	sliderValue = max;
      if(sliderValue < mini)
	sliderValue = mini;
      if((sliderValue - floatValue) > 1.0 ||
	 (sliderValue - floatValue) < - 1.0)
	floatValue = sliderValue;
      arg.setText(value());
    }
  }

  class ButtonChange implements ActionListener{
    public void actionPerformed(ActionEvent e){
      if(e.getSource() == pDotFiveB)
	floatValue += 0.5;
      if(e.getSource() == mDotFiveB)
	floatValue -= 0.5;
      if(e.getSource() == pDotOneB){
	int i = (int)(floatValue*10);
	i += 1;
	floatValue = ((float)i)/10;
      }
      if(e.getSource() == mDotOneB){
	int i = (int)(floatValue*10);
	i -= 1;
	floatValue = ((float)i)/10;
      }
      if(floatValue > max)
	floatValue = max;
      if(floatValue < mini)
	floatValue = mini;
      arg.setText(value());
      slider.setValue(new Float(floatValue).intValue());
    }
  }

  class ResetValue implements ActionListener{
    public void actionPerformed(ActionEvent e){
      floatValue = defaultV;
      slider.setValue(new Float(floatValue).intValue());
      arg.setText(value());
    }
  }

}







