package ORG.ebiss.sara.client;

import java.awt.*;
import ORG.ebiss.sara.server.UIObject;
import ORG.ebiss.sara.server.IntegerParam;

import java.awt.event.*;
import com.sun.java.swing.*;
import com.sun.java.swing.event.*;
import com.sun.java.swing.border.*;

/**
 * Integer parameter Input.
 */

class IntegerInputParam extends ParamView {
  private JTextField arg;     // player's input component
  private JSlider slider;
  private int mini, max, intValue, defaultV;
  private JButton resetB;
  
  IntegerInputParam (UIObject s) {
    super(s);   // the super class constructor 

    IntegerParam p =  (IntegerParam)s;
   
    if (defaultVal != null) {
      try{
	defaultV = Integer.parseInt(defaultVal);
      }catch(NumberFormatException ex){
	System.out.println(ex);
      }
      arg = new JTextField(defaultVal,15);
      intValue = defaultV;
    } else {
      arg = new JTextField("0", 15);
      defaultV = 0;
      intValue = defaultV;
    }

    mini = p.getMin();
    max = p.getMax();

    slider =
      new JSlider(JSlider.HORIZONTAL, mini, max, intValue);
    slider.addChangeListener(new SlideChange());

    arg.addActionListener(new TextFieldInput());
    
    resetB = new JButton("RESET");
    resetB.addActionListener(new ResetValue());

    GridBagLayout gb = new GridBagLayout();
    setLayout(gb);
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = GridBagConstraints.BOTH;

    if (argName != null){
      argName.setFont(Client.FONT);
      gbc.gridwidth  = GridBagConstraints.REMAINDER;
      gb.setConstraints(argName, gbc);
      add(argName);
    }

    gb.setConstraints(slider, gbc);
    add(slider);

    gbc.gridwidth  = GridBagConstraints.RELATIVE;
    gbc.weightx = 1.0;
    gb.setConstraints(arg, gbc);
    add(arg);
    
    JPanel pin = new JPanel();
    JPanel pout = new JPanel();
    pout.setLayout(new BorderLayout(0, 5));
    pin.add(resetB);
    pout.add(pin, BorderLayout.CENTER);

    gbc.weightx = 0.1;
    gbc.gridwidth  = GridBagConstraints.REMAINDER;
    gb.setConstraints(resetB, gbc);
    add(pout);

    value = Integer.toString(intValue); 
  }

  public boolean valid() {
    try {  // check the validy of the number
      Integer.valueOf(value()); 
    } catch (java.lang.NumberFormatException ne) {
      Log.warning("ParamView.valid: NumberFormatException: " + ne);
      return false;
    } catch (NullPointerException e) {
      Log.warning("ParamView.valid: NullPointerException: " + e);
      return false;
    }
    return true;
  }

  /* Get the argument. Overhides super's method. */
  public String value() {
    value = Integer.toString(intValue); 
    return value;
  }

  public void setEnabled(boolean b){
    arg.setEditable(b);
    slider.setEnabled(b);
  }

  class TextFieldInput implements ActionListener{
    public void actionPerformed(ActionEvent e){
      try{
	intValue = Integer.parseInt(arg.getText());
      }catch(NumberFormatException ex){
	System.out.println(ex);
	arg.setText(value());
	return;
      }
      if(intValue > max)
	intValue = max;
      if(intValue < mini)
	intValue = mini;
      slider.setValue(intValue);
      arg.setText(value());
    }
  }
  
  class SlideChange implements ChangeListener{
    public void stateChanged(ChangeEvent e){
      intValue = slider.getValue();
      arg.setText(value());
    }
  }
  
  class ResetValue implements ActionListener{
    public void actionPerformed(ActionEvent e){
      intValue = defaultV;
      slider.setValue(intValue);
      arg.setText(value());
    }
  }
}





