package ORG.ebiss.sara.client;

import java.awt.*;
import ORG.ebiss.sara.server.*;
import java.util.Enumeration;
import java.util.Vector;
import com.sun.java.swing.*;
import java.awt.event.*;


/**
 * Interaction Frame.
 * 	Handles player's interaction. 
 */

public class InteractFrame extends JFrame{
  public static final int IPANEL_COLS = 30;    // TextArea, TextField
  public static final int IPANEL_ROWS = 30;    // TextArea
  public final int FRAME_WIDTH = 300;
  public final int FRAME_HEIGHT = 200;

  private final Color _color = Color.yellow;
  
  private Client top; 
  private Vector params; // list of parameters for each interaction
  private PlayersAction pAction; // player's action object
  private SimValueAdapter adapter;  // interface to server
  
  private Container basePane;
  private JScrollPane scroll;
  private JButton submitButton, cancelButton;
  private JPanel panelOut, panelIn, pane;
  private ListPanel centerPanel;
  
  private String whatThisIs = "";
  
  InteractFrame(SimValueAdapter adapter) {
    super("Interact Frame");
    this.adapter = adapter;
    setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    pane = new JPanel();
    pane.setLayout(new BorderLayout());
    //    scroll = new JScrollPane(pane);
    basePane = getContentPane();
    basePane.setLayout(new GridLayout(0,1));
    //    basePane.add(scroll);
    basePane.add(pane);
  }      
  
  /* setClient: Client -> void
   *	Sets the current top level client. I need this to
   *	call chat in that client.
   */
  
  public void setClient(Client top) { this.top = top; }
  
  private boolean lock = false;  // synchonization lock to iFrame access
  
  /* Reset current interaction
   */
  
  public void resetInteraction () {
    lock = false;
    //    notify();  // process pending interactions
    setVisible(false);
    dispose();
  }
  
  /* interact: Interaction -> void
   *
   *	Make and display the player's interaction panel.
   *	Each parameter uses a separate panel, and handles
   *	the player's input of the argument and the 
   *	"submit/cancel" button actions.
   */
  
  public synchronized void interact(Interaction i) {
    if(i.tag() == Interaction.TAG_DIRECT_INTERACTION){
      whatThisIs = whatThisIs.concat("Interaction :");
    }
    else
      whatThisIs = whatThisIs.concat("Action :");
    
    params = new Vector();
    centerPanel = new ListPanel();
    
    // Now the ordinary interaction.
    // First, the header (sender, reply, comments, etc.)
    for (Enumeration infos = i.elements(); infos.hasMoreElements(); ) {  
      UIObject info = (UIObject)infos.nextElement();
      centerPanel.addList(Util.obj2view(info,IPANEL_COLS));
      whatThisIs = whatThisIs.concat(info.toString() + ":");
    }
    
    // For each parameter, display it on the interaction
    // panel, and wait for player's input, if needed.
    try {
      pAction = i.createAction();   // the interaction part
    } catch (Exception ioe) {
      Log.error("interaction.createAction: " + ioe.getMessage());
      return;
    }
    
    for (Enumeration paramEnum = pAction.elements(); 
	 paramEnum.hasMoreElements();) {
      UIObject param  = 
	(UIObject)paramEnum.nextElement();
      switch(param.getType()) {
      case UIObject.IS_A_LABEL: 
	JLabel label = new JLabel(param.toString()); 
	label.setFont(Client.FONT);
	centerPanel.addList(label);
	whatThisIs = whatThisIs.concat(param.toString() + ":");
	break;
      case UIObject.IS_A_TEXT: 
	JTextArea ta = new JTextArea(param.toString()); 
	ta.setEditable(false);
	ta.setFont(Client.FONT);
	centerPanel.addList(ta);
	whatThisIs = whatThisIs.concat(param.toString() + ":");
	break;
      case UIObject.IS_A_SHORT_STRING: 
      case UIObject.IS_A_SYMBOL: 
	TextFieldParam v3 = new TextFieldParam(param); 
	params.addElement(v3);
	centerPanel.addList(v3);
	break;
      case UIObject.IS_A_LONG_STRING: 
	TextAreaParam v4 = new TextAreaParam(param); 
	params.addElement(v4);
	centerPanel.addList(v4);
	break;
      case UIObject.IS_A_CHOICE: 
	RadioButtonParam v5 = new RadioButtonParam(param); 
	params.addElement(v5);
	centerPanel.addList(v5);
	break;
      case UIObject.IS_A_CHECKBOX: 
	CheckboxParam v6 = new CheckboxParam((UICheckBox)param); 
	params.addElement(v6);
	centerPanel.addList(v6);
	break;
      case UIObject.IS_A_INTEGER_PARAM: 
	IntegerInputParam v7 = new IntegerInputParam((IntegerParam)param); 
	params.addElement(v7);
	centerPanel.addList(v7);
		break;
      case UIObject.IS_A_FLOAT_PARAM: 
	FloatInputParam v8 = new FloatInputParam((FloatParam)param); 
	params.addElement(v8);
	centerPanel.addList(v8);
	break;
      case UIObject.IS_A_SLIDER:
	Slidebar v9 = new Slidebar((Slider)param);
	params.addElement(v9);
	centerPanel.addList(v9);
	break;
      default:
	Log.error("InteractFrame.interact: unknown parameter type" +
		  param.getType());
      }
    }
    
    centerPanel.validate();
    scroll = new JScrollPane(centerPanel);
    //    pane.add(centerPanel, BorderLayout.CENTER);
    pane.add(scroll, BorderLayout.CENTER);

    
    /* Must not add Cancel Button 
     *  When Intearaction is not Player's Action
     */
    
    panelOut = new JPanel();
    panelIn = new JPanel();
    submitButton = new JButton("SUBMIT");
    submitButton.addActionListener(new ButtonListener());
    panelIn.add(submitButton);
    
    cancelButton = new JButton("CANCEL");
    cancelButton.addActionListener(new ButtonListener());
    panelIn.add(cancelButton);
    
    panelOut.setLayout(new BorderLayout(0,5));
    panelOut.add("Center", panelIn);
    pane.add(panelOut, BorderLayout.SOUTH);
    
    validate();

    int width, height;
    width = getPreferredSize().width;
    height = getPreferredSize().height;
    if(width < FRAME_WIDTH)
      width = FRAME_WIDTH;
    if(height < FRAME_HEIGHT)
      height = FRAME_HEIGHT;
    else if(height > 500)
      height = 500;
    setSize(width, height);
    setVisible(true);

    Point point = top.getFrameLocationToBe(getSize().width, getSize().height);
    setLocation(point.x, point.y);

    setBackground(_color);
  }
  
  private synchronized void keepThisFrame(){
    panelIn.removeAll();
    centerPanel.setEnabled(false);
    cancelButton = new JButton("CLOSE");
    cancelButton.addActionListener(new ButtonListener());
    panelIn.add(cancelButton);
    panelIn.validate();
    top.keepInteractFrame(whatThisIs, this);
  }
  
  class ButtonListener implements ActionListener{
    public void actionPerformed(ActionEvent e){
      if(e.getSource() == submitButton){
	if(submitInteraction()) {  // submit ok.
	  /* To Keep Interaction */
	  keepThisFrame();
	  resetInteraction();
	  System.out.println("Interact was submited successfully");
	}else{
	  System.out.println("Interact was not able to be submited");
	  warn();
	}
      }
      if(e.getSource() == cancelButton){      
	resetInteraction();	
	System.out.println("Interact is canceled");
      }
    }
    private void warn(){
      String[] msg = {"Can't Submit !!",
		      "Check your Input or",
		      "Press CANCEL on Frame"};
      JOptionPane.showMessageDialog(pane, msg, "Warning",
				    JOptionPane.WARNING_MESSAGE);
    }
  }
  
  /* submitInteraction:
   *	Check the validy of parameters, and send to server.
   */
  private boolean submitInteraction(){
    /* The first loop checks the validity of
     * the arguments. The second loop sets the value of each parameter
     * (UIOject). I need this second loop because 
     * the server doesn't like updating a previously set argument.
     */
    for(Enumeration en = params.elements(); en.hasMoreElements(); ) {
      ParamView pv = (ParamView)en.nextElement();
	    if (! pv.valid()) {
	      return(false);
	    }
    }
    
    /* All arguments are ok at this point. Set them and send to server.
     * I need a brand new enumeration here. 
     */
    for(Enumeration en1 = params.elements(); en1.hasMoreElements(); ) {
      ParamView pv1 = (ParamView)en1.nextElement();
      pv1.setValue();
      whatThisIs = whatThisIs.concat(pv1.toString() + ":");
      pv1.setEnabled(false);
    }
    try {
      adapter.execute(pAction);
      System.out.println(pAction);
    } catch (Exception ioe) {
      Log.error("adapter.execute: " + ioe.getMessage());
      return(false);
    }
    return(true);
  }
  
  public Insets getInsets() {
    return new Insets(50, 20, 20, 20);
  }
  
}











