/* Util.java:
 *	Utility programs.
 */

package ORG.ebiss.sara.client;

import java.awt.*;
import java.util.Enumeration;
import COM.objectspace.jgl.Array;
import ORG.ebiss.sara.server.UIObject;
import ORG.ebiss.sara.server.CompositeUIObject;

import com.sun.java.swing.*;


  /*********************** UTILITIES ****************************/


public class Util {

  static final int INDENT = 2;  // indentation 

  /* stripnl: String -> String
   *	Strip new lines and nulls from arg string.
   */
  public static String stripnl(String s) {
    String out = "";
    char c;

    for (int i = 0; i < s.length(); i++) {
      c = s.charAt(i);
      if (c != '\n' && c != '\r' && c != '\0') out = out + c; 
    }
    return out;
  }

  /* indent: Int -> String
   *	Returns a string with N spaces. 
   */
  public static String indent(int n ) {
    String out = "";
    for (int i = 0; i < n; i++) {  // veeery inneficient!
      out = out + ' ';
    }
    return out;
  }

  /* cat: String[] -> String
   *	Concatenate a vector V of strings into a single string
   *	with each input string separated by a new line.
   */
  public static String cat (String v[]) {
    String s = "";
    for (int i = 0; i < v.length -1; i++) {
      s = s + v[i] + "\n";
    }
    return s + v[v.length -1];
  }

  /* cat: String[] x Int -> String
   *	Same as cat above, but each line is indented by INDENT spaces. 
   */
  public static String cat (String v[], int indent) {
    if ( indent <= 0 ) return cat(v);  // avoid empty array. 
    char[] indv = new char[indent];
    for (int i = 0;  i < indent; i++) { 
      // initialize character array
      indv[i] = ' ';
    }
    String ind = new String(indv);
    String out = "";
    for (int i = 0; i < v.length -1; i++) {
      out = out + ind + v[i] + "\n";
    }
    return out + ind + v[v.length -1];  // don't put newline at the end
  }    
    

  /* splitText: String x Int -> String
   *	Split the input INP into strings with length =
   *	COLS, indented with INDENT number of spaces. 
   *	The output string is separated by new lines. 
   */

  public static String[] splitText (String inp, int cols) {
    String s = stripnl(inp);      // strip new lines and nulls.
    int length = s.length();
    int rest = length % cols;  // remaining characters
    int rows = rest == 0? length/cols : length/cols +1; // plain rows
    String[] v = new String[rows];

    // First, the plain rows.
    for (int i=0; i < (rest==0? rows : rows-1); i++) {
      v[i] = s.substring(cols * i , cols * i + cols);
    }
    if (rest != 0) {
      v[rows-1] = s.substring(length - rest, length);
    }
    return v;
  }

  /* splitStr: String -> Array[String]
   *	Splits a string into an array of lines. Line terminators
   *	('\n' and '\r') are stripped.
   */

  public static Array str2lines (String s) {
    Array lines = new Array();
    int start, i;
    for (start = 0, i = 0; i < s.length(); ) {
      char c = s.charAt(i);
      if (c != '\n' && c != '\r') { i++; continue; }
      try {  // got a newline
	lines.add(s.substring(start, i)); 
      } catch (IndexOutOfBoundsException e) {
	System.err.println("str2txt: IndexOutOfBound in \""+ s + "\"");
	break;
      } 
      start = ++i; // skip line terminator
    }
    lines.add(s.substring(start));  // remainder of input string
    return lines;
  }


  /* splitLabel: String -> Component
   *	Checks whether the argument string fits the cols;
   *	returns a label if it does, and a panel of two or
   *	more labels otherwise. Labels are aligned LEFT.
   */

  public static Component splitLabel(String s, int cols) {
    int length = s.length();
    int rows = length / cols;  // plain rows
    int remi = length % cols;  // remaing characters

    if (rows == 0) { return new JLabel(s); }
    JPanel p = new JPanel();
    p.setLayout(new GridLayout(0,1)); // labels in a column
    for (int i=0; i < rows; i++) {
      p.add(new JLabel(s.substring(cols * i , cols * i + cols),
		       JLabel.LEFT));
    }
    if ( remi > 0 ) {
      p.add(new JLabel(s.substring(length - remi,length),JLabel.LEFT)); 
    }
    p.validate();
    JScrollPane scroll =
      new JScrollPane(p,
		      ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
		      ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    return scroll;
  }


  /* obj2text: UIObject x int x int -> String
   *	Compose an object into a single text separated into
   *	lines, each with COLS characters. Lines from the second 
   *	are indented with optional INDENT number of spaces. 
   */

  public static String obj2text(UIObject o, int cols) {
    return obj2text(o, cols, -1);  // call obj2text without indentation
  }

  public static String obj2text(UIObject o, 
			  int cols, int indentLevel) {
    int descrIndent;  // indentation for description
    int valueIndent;  // indentation for value
    if (indentLevel < 0) {
      descrIndent = 0; 
      valueIndent = 0; 
    } else {
      descrIndent  = indentLevel * INDENT;
      valueIndent = descrIndent + INDENT;
    }

    String objDescr = o.getDescription();
    // Log.log("objDescr=" + objDescr + ".");
    String descr = "+ ";
    if (objDescr != "")  descr = descr + objDescr + " = ";
    String text = "";

    switch(o.getType()) {
    case UIObject.IS_A_LABEL: 
      text = o.toString();
      break;

    case UIObject.IS_A_INTEGER: 
    case UIObject.IS_A_FLOAT: 
    case UIObject.IS_A_INTEGER_PARAM: 
    case UIObject.IS_A_FLOAT_PARAM: 
    case UIObject.IS_A_SLIDER:
      text = indent(descrIndent) + descr + o.toString();
      break;

    case UIObject.IS_A_SHORT_STRING: 
    case UIObject.IS_A_SYMBOL: 
      String s = o.toString();
      text = indent(descrIndent) + descr + s;
      break;

    case UIObject.IS_A_TEXT: 
    case UIObject.IS_A_LONG_STRING: 
      String txt = o.toString();
      if (txt.length() <= cols) {
	text = txt;
      } else { 
	text = cat(splitText(txt, cols), valueIndent);
      }
      break;
/*
      String txt = o.toString();
      if (descrIndent + descr.length() + txt.length() <= cols) {
	text = indent(descrIndent) + descr + txt;
      } else { 
	text = cat(splitText(descr,cols),descrIndent) + '\n' +
	  cat(splitText(txt, cols), valueIndent);
      }
      break;
*/

    case UIObject.IS_A_CHOICE: 
      text = descr + o.toString(); // the selected value
      // text = o.toString(); 
      break;
    
    case UIObject.IS_COMPOSITE: 
      /*
	if (objDescr != "")
	text = cat(splitText(descr,cols), descrIndent);
      */
      Enumeration elems = ((CompositeUIObject) o).elements();
      while (elems.hasMoreElements()) {
	UIObject elem =(UIObject)elems.nextElement();
	text = text + "\n" + obj2text(elem,cols,indentLevel +1);
      }	
      break;
      
    default: 
      text = descr + "unknown object type";
      break;
    } 
    return(text);
  }


  /* obj2view: SelfdescribingObject -> Component
   *	Make an ui component of an object O to be 
   *	put in a container COLS columns wide. 
   */ 

  public static Component obj2view (UIObject o, 
				    int cols) {
    String txt = obj2text(o,cols,0);
    JPanel p;
    JLabel label;

    switch(o.getType()) {
    case UIObject.IS_A_LABEL: 
    case UIObject.IS_A_INTEGER: 
    case UIObject.IS_A_FLOAT: 
    case UIObject.IS_A_INTEGER_PARAM: 
    case UIObject.IS_A_FLOAT_PARAM: 
    case UIObject.IS_A_SHORT_STRING: 
    case UIObject.IS_A_SYMBOL: 
    case UIObject.IS_A_CHOICE: 
      label = new JLabel(txt);
      label.setFont(Client.FONT);
      return label;

    case UIObject.IS_A_TEXT: 
    case UIObject.IS_A_LONG_STRING: 
    case UIObject.IS_COMPOSITE: 
      p = new JPanel();
      p.setLayout(new GridLayout(0,1));
      Array lines = str2lines(txt);
      if (lines.size() == 1){
	label = new JLabel(txt);
	label.setFont(Client.FONT);
	return label;
      }
      for ( Enumeration en = lines.elements(); en.hasMoreElements(); ) {
	label = new JLabel((String)en.nextElement());
	label.setFont(Client.FONT);
	p.add(label);
      }
      JScrollPane scroll =
	new JScrollPane(p,
			ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS,
			ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
      return scroll;
      
    default: 
      System.err.println("obj2view: unkown object type: " + o.getType());
      label = new JLabel("obj2view: unkown object type: " + o.getType());
      label.setFont(Client.FONT);
      return label;
    }
  }
}

