package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import ORG.ebiss.sara.parser.*;

/**
 * <code>Code</code>֥ɤɽ륯饹Ǥ.
 * 
 * @see		ORG.ebiss.sara.interpreter.Interpreter
 * @version	1.0 Sep 1997
 * @author	TOGAWA Atsushi
 */
public abstract class Code implements Serializable {

  public Code() {}

  public Code(Node x) { node = x; }

  protected void duplicateDefault() throws EvaluationException {
    stop("Duplicate default targets");
  }

  protected void misplacedBreak() throws EvaluationException {
    stop("No enclosing while, do, for, or switch statement");
  }

  protected void misplacedContinue() throws EvaluationException {
    stop("No enclosing while, do, or for statement");
  }

  protected void misplacedReturn() throws EvaluationException {
    stop("No enclosing procedure definition");
  }

  protected void undefinedLabel(String label) throws EvaluationException {
    stop("Undefined label \"" + label + "\"");
  }

  protected void noMatchingMethod() throws EvaluationException {
    stop("No matching method is found");
  }

  protected void noMatchingMethod(String name) throws EvaluationException {
    stop("No matching method is found: \"" + name + "\"");
  }

  protected void noMatchingProcedure() throws EvaluationException {
    stop("No matching procedure is found");
  }

  protected void undefinedField(Value o, String n) throws EvaluationException {
    stop("Undefined field \"" + n + "\"");
  }

  protected void undefinedVariable(String name) throws EvaluationException {
    stop("Undefined variable \"" + name + "\"");
  }

  protected void misplacedThis() throws EvaluationException {
    stop("Misplaced this");
  }

  protected void javaClassNotFound(String name) throws EvaluationException {
    stop("Java class \"" + name + "\" not found");
  }

  protected void domainError(Value given)
       throws EvaluationException
  {
    stop("domain error");
  }

  protected void domainError(Value given, String expected)
       throws EvaluationException
  {
    stop(expected + " is expected, but the given value is \"" +
	 given.getPrintableRepresentation() + "\"");
  }

  protected void invalidOperation(int opcode, Value value)
       throws EvaluationException
  {
    switch (opcode) {
    case UnaryOpExpression.PLUS:
      break;
    case UnaryOpExpression.MINUS:
      break;
    case UnaryOpExpression.NOT:
      break;
    case UnaryOpExpression.CONDITIONAL_NOT:
      break;
    case UnaryOpExpression.INCREMENT:
      break;
    case UnaryOpExpression.DECREMENT:
      break;
    default:
      throw new RuntimeException();
    }
    // XXX: ȤåϤޤ
    stop("invalid operation");
  }

  protected void invalidOperation(int opcode, Value left, Value right)
       throws EvaluationException
  {
    String	op = null;

    switch (opcode) {
    case BinaryOpExpression.MULTIPLY:
      op = "*";
      break;
    case BinaryOpExpression.DIVIDE:
      op = "/";
      break;
    case BinaryOpExpression.REMAINDER:
      op = "%";
      break;
    case BinaryOpExpression.PLUS:
      op = "+";
      break;
    case BinaryOpExpression.MINUS:
      op = "-";
      break;
    case BinaryOpExpression.LEFT_SHIFT:
      op = "<<";
      break;
    case BinaryOpExpression.RIGHT_SIGNED_SHIFT:
      op = ">>";
      break;
    case BinaryOpExpression.RIGHT_UNSIGNED_SHIFT:
      op = ">>>";
      break;
    case BinaryOpExpression.AND:
      op = "&";
      break;
    case BinaryOpExpression.EXCLUSIVE_OR:
      op = "^";
      break;
    case BinaryOpExpression.INCLUSIVE_OR:
      op = "|";
      break;
    case BinaryOpExpression.CONDITIONAL_AND:
      op = "&&";
      break;
    case BinaryOpExpression.CONDITIONAL_OR:
      op = "||";
      break;
    case BinaryOpExpression.EQUAL:
      op = "==";
      break;
    case BinaryOpExpression.NOT_EQUAL:
      op = "!=";
      break;
    case BinaryOpExpression.LESS:
      op = "<";
      break;
    case BinaryOpExpression.GREATER:
      op = ">";
      break;
    case BinaryOpExpression.LESS_EQUAL:
      op = "<=";
      break;
    case BinaryOpExpression.GREATER_EQUAL:
      op = ">=";
      break;
    default:
      throw new RuntimeException();
    }
    // XXX: ȤåϤޤ
    stop("invalid operation");
  }

  protected void failedToJoin(Value target, Value operand)
       throws EvaluationException
  {
    stop("failed to join objects");
  }

  protected void failedToDefineField(Value target, String[] modifiers,
				     String name, Value value)
       throws EvaluationException
  {
    stop("failed to define field \"" + name + "\"");
  }

  protected void failedToWriteField(Value target, String name, Value value)
       throws EvaluationException
  {
    stop("failed to write into field \"" + name + "\"");
  }

  protected void failedToDefineMethod(Value target,
				      String[] modifiers, String name)
       throws EvaluationException
  {
    stop("failed to define method \"" + name + "\"");
  }

  protected void stop(String message) throws EvaluationException {
    throw new EvaluationException(message);
  }

  public String getLocationString() {
    if (node == null) return "<console>:0:0";
    return node.getLocationString();
  }

  public String toString() {
    StringBuffer result = new StringBuffer();
    Class theClass = getClass();
    String name = theClass.getName();
    result.append(System.identityHashCode(this) + ":" +
		  name.substring(name.lastIndexOf('.') + 1) + "(");
    Field[] fields = theClass.getFields();
    for (int i = 0; i < fields.length; ++i) {
      if (i > 0) result.append(", ");
      result.append(fields[i].getName());
      try {
	Object value = fields[i].get(this);
	result.append('=');
	result.append(fieldValueToString(value));
      } catch (IllegalAccessException iae) {
      }
    }
    result.append(")");

    return result.toString();
  }

  protected String fieldValueToString(Object value) {
    StringBuffer result = new StringBuffer();

    if (value == null) {
      result.append("null");
    } else if (value.getClass().isArray()) {
      result.append("(");
      int size = Array.getLength(value);
      for (int i = 0; i < size; ++i) {
	if (i > 0) result.append(", ");
	result.append(fieldValueToString(Array.get(value, i)));
      }
      result.append(")");
    } else if (value instanceof Code) {
      result.append(System.identityHashCode(value));
    } else if (value instanceof Node) {
      result.append(((Node)value).toString(0, 0));
    } else {
      result.append(value.toString());
    }
    return result.toString();
  }

  /**
   * ΥɤͳȤʤäݹʸ.  ¹Ի顼ȯ˥
   * ץΰ֤뤿Ѥ.
   *
   * <p>⤷Ŭڤʹʸڤ¸ߤʤ<code>null</code>Ǥ.
   *
   * @see ORG.ebiss.sara.Node
   */
  public Node	node;

  /**
   * execute᥽åɤϤΥɤ¹Ԥ᥽åɤǤ.
   * ŪʽƤϥ֥饹ˤä.
   * 
   * @param interpreter	Υɤ¹ԤƤ륤󥿥ץ꥿
   * @return		˼¹Ԥ٤
   * @exception		ORG.ebiss.sara.interpreter.EvaluationException
   *			¹˲餫Υ顼Ф줿Ȥȯ.
   */
  public abstract Code execute(Interpreter interpreter)
       throws EvaluationException;
}
