package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;

public class DataSlot extends Slot {
  private boolean	_isInheritSlot;
  protected Value	_value;

  public DataSlot(SpdValue container, String[] modifiers,
		  String name, Value value)
       throws InvalidOperationException
  {
    super(container, modifiers, name);
    _value = value;
    _isInheritSlot = hasModifier("inherit");
    if (_isInheritSlot && !(value instanceof SpdValue)) {
      throw new InvalidOperationException();
    }
  }

  public Slot copy(SpdValue destination) {
    try {
      return destination.createDataSlot(_modifiers, _name, _value);
    } catch (InvalidOperationException uoe) {
      uoe.printStackTrace();
      throw new RuntimeException();
    }
  }

  public Slot freshCopy(SpdValue destination) {
    try {
      return destination.createDataSlot(_modifiers, _name,
					isInheritSlot() ?
					_value : SpdValue.NULL);
    } catch (InvalidOperationException uoe) {
      uoe.printStackTrace();
      throw new RuntimeException();
    }
  }

  public void enter(SpdValue object) {
    if (_isInheritSlot) object.addParent(_name);
  }

  public void leave(SpdValue object) {
    if (_isInheritSlot) object.removeParent(_name);
  }

  public boolean isInheritSlot() { return _isInheritSlot; }

  public Object getPrintableRepresentation(final PrintingContext context) {
    return new Object() {
      Object valuePrinting = _value.getPrintableRepresentation(context);
      public String toString() {
	StringBuffer buffer = new StringBuffer();
	for (Enumeration i = modifiers(); i.hasMoreElements(); ) {
	  buffer.append(i.nextElement());
	  buffer.append(' ');
	}
	buffer.append(name());
	buffer.append(" = ");
	buffer.append(valuePrinting);
	buffer.append(";");
	return buffer.toString();
      }
    };
  }

  public Value read(SpdValue receiver) { return _value; }

  public void write(SpdValue receiver, Value value)
       throws InvalidOperationException
  {
    if (_isInheritSlot && !(value instanceof SpdValue)) {
      throw new InvalidOperationException();
    }
    _value = value;
  }
}
