package ORG.ebiss.sara.interpreter;

import java.util.*;
import COM.objectspace.jgl.*;

public class ImportedSlot extends Slot implements ObservableSlot {
  private boolean	_visible;
  private String	_label;
  private SpdValue	_receiver;
  private Slot		_master;

  ImportedSlot(SpdValue parent, String name, SpdValue object, Slot master)
       throws InvalidOperationException
  {
    super(parent, SpdValue.noModifiers, name);
    _visible = true;
    _label = name;
    _receiver = object;
    while (master instanceof ImportedSlot) {
      master = ((ImportedSlot)master).getMaster();
    }
    _master = master;
  }

  public Slot copy(SpdValue destination) {
    try {
      return new ImportedSlot(destination, name(), _receiver, _master);
    } catch (InvalidOperationException uoe) {
      uoe.printStackTrace();
      throw new RuntimeException();
    }
  }

  public Slot getMaster() { return _master; }

  public boolean visible() { return _visible; }

  public void visible(boolean newValue) { _visible = newValue; }

  public String label() { return _label; }

  public void label(String label) { _label = label; }

  public Object getPrintableRepresentation(final PrintingContext context) {
    return new Object() {
      Object valuePrinting = _master.getPrintableRepresentation(context);
      public String toString() {
	StringBuffer buffer = new StringBuffer();
	for (Enumeration i = modifiers(); i.hasMoreElements(); ) {
	  buffer.append(i.nextElement());
	  buffer.append(' ');
	}
	buffer.append(name());
	buffer.append(" = ");
	buffer.append(valuePrinting);
	buffer.append(";");
	return buffer.toString();
      }
    };
  }

  public Value read(SpdValue receiver) {
    return _master.read(_receiver);
  }

  public void write(SpdValue receiver, Value value)
       throws InvalidOperationException
  {
    _master.write(_receiver, value);
  }
}
