package ORG.ebiss.sara.interpreter;

import java.util.*;
import ORG.ebiss.sara.parser.*;

/**
 * 
 * 
 * @version	1.0 Oct 1997
 * @author	TOGAWA Atsushi
 */
public final class IntegerValue extends Value implements Atomic {
  private int	_value;

  public IntegerValue(int value) { _value = value; }

  /*========================================================================
   *	ͤФŪ
   *========================================================================*/

  public boolean equals(Object v) {
    return v != null &&
      v instanceof IntegerValue && ((IntegerValue)v)._value == _value;
  }

  public int hashCode() { return _value; }

  public boolean equalsToLabel(Object v) {
    if (! (v instanceof Integer)) return false;
    return ((Integer)v).intValue() == _value;
  }

  /*========================================================================
   *		黻
   *========================================================================*/

  public Value binaryOp(int opcode, Value operand)
       throws InvalidOperationException, CoerceFailedException
  {
    if (operand instanceof DoubleValue) {
      switch (opcode) {
      case BinaryOpExpression.MULTIPLY:
      case BinaryOpExpression.PLUS:
	return operand.binaryOp(opcode, this);

      case BinaryOpExpression.DIVIDE:
      case BinaryOpExpression.REMAINDER:
      case BinaryOpExpression.MINUS:
	return new DoubleValue(this.doubleValue()).binaryOp(opcode, operand);

      case BinaryOpExpression.LESS:
	return operand.binaryOp(BinaryOpExpression.GREATER, this);

      case BinaryOpExpression.GREATER:
	return operand.binaryOp(BinaryOpExpression.LESS, this);

      case BinaryOpExpression.LESS_EQUAL:
	return operand.binaryOp(BinaryOpExpression.GREATER_EQUAL, this);

      case BinaryOpExpression.GREATER_EQUAL:
	return operand.binaryOp(BinaryOpExpression.LESS_EQUAL, this);

      }
    }

    switch(opcode) {
    case BinaryOpExpression.MULTIPLY:
      return new IntegerValue(_value * operand.intValue());
    case BinaryOpExpression.DIVIDE:
      return new IntegerValue(_value / operand.intValue());
    case BinaryOpExpression.REMAINDER:
      return new IntegerValue(_value % operand.intValue());
    case BinaryOpExpression.PLUS:
      return new IntegerValue(_value + operand.intValue());
    case BinaryOpExpression.MINUS:
      return new IntegerValue(_value - operand.intValue());
    case BinaryOpExpression.LEFT_SHIFT:
      return new IntegerValue(_value << operand.intValue());
    case BinaryOpExpression.RIGHT_SIGNED_SHIFT:
      return new IntegerValue(_value >> operand.intValue());
    case BinaryOpExpression.RIGHT_UNSIGNED_SHIFT:
      return new IntegerValue(_value >>> operand.intValue());
    case BinaryOpExpression.AND:
      return new IntegerValue(_value & operand.intValue());
    case BinaryOpExpression.EXCLUSIVE_OR:
      return new IntegerValue(_value ^ operand.intValue());
    case BinaryOpExpression.INCLUSIVE_OR:
      return new IntegerValue(_value | operand.intValue());
    case BinaryOpExpression.LESS:
      return _value < operand.intValue()
	? BooleanValue.TRUE : BooleanValue.FALSE;
    case BinaryOpExpression.GREATER:
      return _value > operand.intValue()
	? BooleanValue.TRUE : BooleanValue.FALSE;
    case BinaryOpExpression.LESS_EQUAL:
      return _value <= operand.intValue()
	? BooleanValue.TRUE : BooleanValue.FALSE;
    case BinaryOpExpression.GREATER_EQUAL:
      return _value >= operand.intValue()
	? BooleanValue.TRUE : BooleanValue.FALSE;

    default:
      return super.binaryOp(opcode, operand);
    }
  }

  public Value unaryOp(int opcode) throws InvalidOperationException {
    switch(opcode) {
    case UnaryOpExpression.PLUS:
      return this;
    case UnaryOpExpression.MINUS:
      return new IntegerValue(-_value);
    case UnaryOpExpression.NOT:
      return new IntegerValue(~_value);
    default:
      throw new InvalidOperationException();
    }
  }

  /*========================================================================
   *	¾ΥǡؤѴ뤤¾ΥǡѴ
   *========================================================================*/

  public int intValue() { return _value; }

  public double doubleValue() { return 0.0 + _value; }

  public String stringValue() { return Integer.toString(_value); }

  public Object javaValue() { return new Integer(_value); }

  public Class javaClass() { return Integer.TYPE; }

  /*========================================================================
   *	ɽǽʸɽؤѴ
   *========================================================================*/

  public Object getPrintableRepresentation(PrintingContext context) {
    return Integer.toString(_value);
  }
}
