package ORG.ebiss.sara.interpreter;

final class JavaConstructorSlot extends JavaSlot {
  private java.lang.reflect.Constructor[]	_constructors;

  JavaConstructorSlot(java.lang.reflect.Constructor[] constructors) {
    _constructors = constructors;
  }

  public Value read(JavaValue receiver) {
    return new JavaConstructorValue(_constructors);
  }
}

final class JavaConstructorValue extends ApplicableValue {
  java.lang.reflect.Constructor[]	_constructors;

  JavaConstructorValue(java.lang.reflect.Constructor[] constructors) {
    _constructors = constructors;
  }

  public Code apply(Interpreter interpreter, int[] arguments,
		    int resultSlot, Code returnPoint)
       throws EvaluationException
  {
    Object[] args = new Object[arguments.length];
    Class[] classes = new Class[arguments.length];
    for (int i = arguments.length - 1; i >= 0; --i) {
      Value value = interpreter.readRegister(arguments[i]);
      args[i] = value.javaValue();
      classes[i] = value.javaClass();
    }
    int n = _constructors.length;
  checkMethod:
    for (int i = 0; i < n; ++i) {
      java.lang.reflect.Constructor constructor = _constructors[i];
      Class[] paramTypes = constructor.getParameterTypes();
      if (paramTypes.length != arguments.length) continue;
      for (int j = paramTypes.length - 1; j >= 0; --j) {
	if (! paramTypes[j].isAssignableFrom(classes[j])) {
	  continue checkMethod;
	}
      }
      try {
	Object result = constructor.newInstance(args);
	interpreter.writeRegister(resultSlot, new JavaValue(result));
      } catch (Exception e) {
	throw new EvaluationException(e.getMessage());
      }
      return returnPoint;
    }
    throw new EvaluationException("no matching constructor");
  }

  public boolean checkArity(int arity) {
    int n = _constructors.length;
    for (int i = 0; i < n; ++i) {
      java.lang.reflect.Constructor constructor =
	(java.lang.reflect.Constructor)_constructors[i];
      if (constructor.getParameterTypes().length == arity) return true;
    }
    return false;
  }
}
