package ORG.ebiss.sara.interpreter;

import java.util.*;
import java.lang.reflect.InvocationTargetException;

final class JavaMethodSlot extends JavaSlot {
  private Vector	_methods = new Vector();

  public JavaMethodSlot(java.lang.reflect.Method method) {
    _methods.addElement(method);
  }

  public void add(java.lang.reflect.Method method) {
    _methods.addElement(method);
  }

  public Value read(JavaValue receiver) {
    return new MethodValue(receiver);
  }

  class MethodValue extends ApplicableValue {
    public JavaValue	_receiver;

    MethodValue(JavaValue receiver) { _receiver = receiver; }

    public String getPrintableRepresentation() {
      StringBuffer result = new StringBuffer("<java method: " +
					     _receiver.getJavaValue().getClass().getName() + "#" +
					     ((java.lang.reflect.Method)_methods.elementAt(0)).getName());
      int n = _methods.size();
      for (int i = 0; i < n; ++i) {
	if (i > 0) result.append(" | ");
	result.append('(');
	int l = ((java.lang.reflect.Method)_methods.elementAt(i)).
	  getParameterTypes().length;
	for (int j = 0; j < l; ++j) {
	  if (j > 0) result.append(',');
	  result.append('*');
	}
	result.append(')');
      }
      result.append(">");
      return result.toString();
    }

    public Object getPrintableRepresentation(PrintingContext context) {
      return getPrintableRepresentation();
    }

    public Code apply(Interpreter interpreter, int[] arguments,
		      int resultSlot, Code returnPoint)
	 throws EvaluationException
    {
      Value[] args = new Value[arguments.length];
      Class[] convertedClasses = new Class[arguments.length];
      Class[] rawClasses = new Class[arguments.length];
      for (int i = arguments.length - 1; i >= 0; --i) {
	Value value = args[i] = interpreter.readRegister(arguments[i]);
	convertedClasses[i] = value.javaClass();
	rawClasses[i] = value.getClass();
      }
      int n = _methods.size();
    checkMethod:
      for (int i = 0; i < n; ++i) {
	java.lang.reflect.Method method =
	  (java.lang.reflect.Method)_methods.elementAt(i);
	Class[] paramTypes = method.getParameterTypes();
	if (paramTypes.length != arguments.length) continue;
	for (int j = paramTypes.length - 1; j >= 0; --j) {
	  if (! paramTypes[j].isAssignableFrom(convertedClasses[j]) &&
	      ! paramTypes[j].isAssignableFrom(rawClasses[j])) {
	    continue checkMethod;
	  }
	}

	Object[] jargs = new Object[args.length];
	for (int j = paramTypes.length - 1; j >= 0; --j) {
	  if (paramTypes[j].isAssignableFrom(rawClasses[j])) {
	    jargs[j] = args[j];
	  } else {
	    jargs[j] = args[j].javaValue();
	  }
	}

	try {
	  Object java = method.invoke(_receiver.getJavaValue(), jargs);
	  Class type = method.getReturnType();
	  interpreter.writeRegister(resultSlot, JavaValue.unwrap(java, type));
	} catch (IllegalAccessException acce) {
	  acce.printStackTrace();
	  throw new EvaluationException(acce.getClass().getName() + ": " +
					acce.getMessage());
	} catch (InvocationTargetException ite) {
	  Throwable te = ite.getTargetException();
	  te.printStackTrace();
	  throw new EvaluationException(te.getClass().getName() + ": " +
					te.getMessage());
	} catch (IllegalArgumentException arge) {
	  continue checkMethod;
	}
	return returnPoint;
      }
      throw new EvaluationException("no matching method");
    }

    public boolean checkArity(int arity) {
      int n = _methods.size();
      for (int i = 0; i < n; ++i) {
	java.lang.reflect.Method method =
	  (java.lang.reflect.Method)_methods.elementAt(i);
	if (method.getParameterTypes().length == arity) return true;
      }
      return false;
    }
  }
}
