package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;
import java.lang.reflect.InvocationTargetException;

public class JavaValue extends Value {
  protected Object	_object;
  protected Hashtable	_name2value = new Hashtable();

  public static Class valueClass;

  public JavaValue(Class javaClass) {
    _object = javaClass;
    java.lang.reflect.Field[] fields = javaClass.getFields();
    java.lang.reflect.Method[] methods = javaClass.getMethods();
    java.lang.reflect.Constructor[] constructors = javaClass.getConstructors();

    for (int i = 0; i < fields.length; ++i) {
      java.lang.reflect.Field field = fields[i];
      if ((field.getModifiers() & java.lang.reflect.Modifier.STATIC) == 0) {
	continue;
      }
      _name2value.put(field.getName(), new JavaFieldSlot(field));
    }

    for (int i = 0; i < methods.length; ++i) {
      java.lang.reflect.Method method = methods[i];
      if ((method.getModifiers() & java.lang.reflect.Modifier.STATIC) == 0) {
	continue;
      }
      String name = method.getName();
      JavaMethodSlot slot = (JavaMethodSlot)_name2value.get(name);
      if (slot != null) {
	slot.add(method);
      } else {
	_name2value.put(name, new JavaMethodSlot(method));
      }
    }

    _name2value.put("new", new JavaConstructorSlot(constructors));
  }

  public JavaValue(Object javaObject) {
    _object = javaObject;
    Class javaClass = javaObject.getClass();
    java.lang.reflect.Field[] fields = javaClass.getFields();
    for (int i = 0; i < fields.length; ++i) {
      java.lang.reflect.Field field = fields[i];
      if ((field.getModifiers() & java.lang.reflect.Modifier.STATIC) != 0) {
	continue;
      }
      _name2value.put(field.getName(), new JavaFieldSlot(field));
    }

    java.lang.reflect.Method[] methods = javaClass.getMethods();
    int n = methods.length;
    for (int i = 0; i < n; ++i) {
      java.lang.reflect.Method method = methods[i];
      if ((method.getModifiers() & java.lang.reflect.Modifier.STATIC) != 0) {
	continue;
      }
      String name = method.getName();
      JavaMethodSlot slot = (JavaMethodSlot)_name2value.get(name);
      if (slot != null) {
	slot.add(method);
      } else {
	_name2value.put(name, new JavaMethodSlot(method));
      }
    }

    java.lang.reflect.Constructor[] constructors = javaClass.getConstructors();
    _name2value.put("new", new JavaConstructorSlot(constructors));

    _name2value.put("copy", new JavaSlot() {
      public Value read(final JavaValue receiver) {
	return new ApplicableValue() {
	  public Code apply(Interpreter interpreter, int[] arguments,
			    int resultSlot, Code returnPoint)
	    throws EvaluationException
	  {
	    {
	      if (arguments.length != 0) {
		throw new EvaluationException("no matching method");
	      }
	      try {
		interpreter.writeRegister(resultSlot, receiver.copy());
	      } catch (InvalidOperationException uoe) {
		throw new EvaluationException("copy() is not supported");
	      }
	      return returnPoint;
	    }
	  }

	  public boolean checkArity(int arity) { return arity == 0; }
	};
      }
    });
  }

  public Object getJavaValue() { return _object; }

  public boolean equals(Object x) {
    return x != null &&
      x instanceof JavaValue && ((JavaValue)x)._object.equals(_object);
  }

  public int hashCode() { return _object.hashCode(); }

  public static Value unwrap(Object x, Class type) {
    if (type == Void.TYPE) return Value.UNSPECIFIED;
    if (x == null) return Value.NULL;
    if (x instanceof Value) return (Value)x;
    if (type == Integer.TYPE)
      return new IntegerValue(((Integer)x).intValue());
    if (type == Boolean.TYPE)
      return new BooleanValue(((Boolean)x).booleanValue());
    if (type == Double.TYPE)
      return new DoubleValue(((Double)x).doubleValue());
    if (type == Character.TYPE)
      // XXX: not implemented yet
      return new IntegerValue(((Character)x).charValue());
    if (type == Byte.TYPE || type == Short.TYPE || type == Long.TYPE)
      return new IntegerValue(((Number)x).intValue());
    if (type == Float.TYPE)
      return new DoubleValue(((Float)x).doubleValue());
    if (x instanceof String)
      return new StringValue((String)x);
    if (x instanceof SList)
      return ListValue.valueOf((SList)x);
    if (x instanceof Array)
      return VectorValue.valueOf((Array)x);
    return new JavaValue(x);
  }

  /*========================================================================
   *	ͤФŪ
   *========================================================================*/

  public Value copy() throws InvalidOperationException {
    try {
      JavaValue result = (JavaValue)clone();
      result._object = copyObject();
      result._name2value = (Hashtable)_name2value.clone();
      return result;
    } catch (CloneNotSupportedException cnse) {
      throw new InvalidOperationException();
    }
  }

  protected Object copyObject() throws InvalidOperationException {
    if (_object instanceof Container) {
      return ((Container)_object).clone();
    }
    throw new InvalidOperationException();
  }

  /**
   * ֥ȤͿ줿ͤɤ򸡺.
   * ǡפȤ, SPDץˤäξԤΰ㤤
   * ФʤȤ̣롥
   * @param x	Ӥ
   * @return	<code>true</code>
   */
  public boolean equalsToValue(Value x) {
    if (! (x instanceof JavaValue)) return false;
    return _object == ((JavaValue)x)._object;
  }

  /*========================================================================
   *	¾ΥǡؤѴ뤤¾ΥǡѴ
   *========================================================================*/

  public int intValue() throws CoerceFailedException {
    if (_object instanceof Number) return ((Number)_object).intValue();
    return super.intValue();
  }

  public double doubleValue() throws CoerceFailedException {
    if (_object instanceof Number) return ((Number)_object).doubleValue();
    return super.doubleValue();
  }

  public boolean booleanValue() throws CoerceFailedException {
    if (_object instanceof Boolean) return ((Boolean)_object).booleanValue();
    return super.booleanValue();
  }

  public String stringValue() throws CoerceFailedException {
    return _object.toString();
  }

  public Object javaValue() { return _object; }

  public Class javaClass() { return _object.getClass(); }

  /*========================================================================
   *	ɽǽʸɽؤѴ
   *========================================================================*/

  protected DelayedPrinting createDelayedPrinting() {
    return new Printing(this);
  }

  static class Printing extends DelayedPrinting {
    Printing(JavaValue value) { super(value); }

    public String getPrintableRepresentation() {
      return "<java: " + ((JavaValue)_value)._object.toString() + ">";
    }
  }

  /*========================================================================
   *	֤̾ȤƤ
   *========================================================================*/

  public Value read(String name) {
    JavaSlot slot = (JavaSlot)_name2value.get(name);
    if (slot == null) return null;
    return slot.read(this);
  }

  public boolean write(String name, Value value)
       throws InvalidOperationException
  {
    JavaSlot slot = (JavaSlot)_name2value.get(name);
    if (slot == null) return false;
    slot.write(this, value);
    return true;
  }

  public boolean contains(String name) {
    return _name2value.get(name) != null;
  }

  static {
    try {
      valueClass = Class.forName("ORG.ebiss.sara.interpreter.Value");
    } catch (ClassNotFoundException cnfe) {
      cnfe.printStackTrace();
      throw new RuntimeException();
    }
  }
}
