package ORG.ebiss.sara.interpreter;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;

public class SlotEnumeration implements Cloneable, Serializable, Enumeration {
  public SpdValue	receiver;
  public String	name;
  public Slot		cache;
  public Vector	lookupQueue = new Vector();

  public SlotEnumeration(SpdValue _receiver, String _name) {
    name = _name;
    receiver = _receiver;
    if (_receiver != SpdValue._object) {
      lookupQueue.addElement(SpdValue._object);
    }
    lookupQueue.addElement(_receiver);
  }

  public Object clone() {
    try {
      SlotEnumeration result = (SlotEnumeration)super.clone();
      result.lookupQueue = (Vector)lookupQueue.clone();
      return result;
    } catch (CloneNotSupportedException cnse) {
      cnse.printStackTrace();
      throw new RuntimeException();
    }
  }

  public String getName() { return name; }

  public void skip() {
    lookup();
    cache = null;
  }

  protected void lookup() {
    SpdValue object = (SpdValue)lookupQueue.lastElement();
    lookupQueue.removeElementAt(lookupQueue.size() - 1);
    for (Enumeration e = object.parents(); e.hasMoreElements(); ) {
      SpdValue parent = (SpdValue)
	((Slot)object._slots.get((String)e.nextElement())).read(receiver);
      lookupQueue.addElement(parent);
    }
    cache = (Slot)object._slots.get(name);
  }

  private void fillCache() {
    while (cache == null) {
      if (lookupQueue.isEmpty()) return;
      lookup();
    }
  }

  public boolean hasMoreElements() {
    fillCache();
    return cache != null;
  }

  public Object nextElement() throws NoSuchElementException {
    fillCache();
    if (cache == null) throw new NoSuchElementException();
    Object result = cache;
    cache = null;
    return result;
  }
}
