package ORG.ebiss.sara.interpreter;

import java.util.*;
import ORG.ebiss.sara.parser.Node;
import ORG.ebiss.sara.parser.Symbol;
import ORG.ebiss.sara.parser.UnaryOpExpression;
import ORG.ebiss.sara.parser.BinaryOpExpression;
import ORG.ebiss.sara.parser.ConditionalOpExpression;
import ORG.ebiss.sara.parser.SwitchBlockStatements;
import ORG.ebiss.sara.parser.ProcedureConstruction;
import ORG.ebiss.sara.parser.ParenthesizedExpression;
import ORG.ebiss.sara.parser.IncOrDecExpression;
import COM.objectspace.jgl.*;

/**
 * <code>Translator</code>ݹʸ(<code>Node</code>)֥
 * (<code>Code</code>)ؤѴǤ.
 *
 * @see		ORG.ebiss.sara.parser.Node
 * @see		ORG.ebiss.sara.parser.Node#accept(NodeVisitor)
 * @see		ORG.ebiss.sara.interpreter.Code
 * @version	1.0 Sep 1997
 * @author	TOGAWA Atsushi
 */
public final
class Translator extends ORG.ebiss.sara.parser.PartialNodeVisitor {
  /**
   * Ϳ줿ݹʸڤ֥ɤ.
   *
   * @param node	Ѵץιʸ
   * @param successor	줿ɤ¹Ԥλܤ
   * 			. nullξ, 󥿥ץ꥿λ.
   * @param registerLowLimit	줿ɤѹƤ褤쥸β
   * @param registerSetSize	ߤΥ쥸礭
   * @param definedLabels	ѤΥ٥
   * @param noLabelBreakIsValid	٥Τʤ<code>break</code>ʸ¹Բǽ
   *				ʸ̮Ǥ뤫ɤ򼨤
   * @param noLabelContinueIsValid
   *				٥Τʤ<code>continue</code>ʸ¹Բǽ
   *				ʸ̮Ǥ뤫ɤ򼨤
   * @param inProcedure	⤷³ΤѴȤ<code>true</code>.
   *			Ǥʤ<code>false</code>
   * @return		줿ɤΥȥݥȤɬפʥ쥸ο
   */
  public static Executable translate(Node node, Code successor,
				     int registerLowLimit,
				     int registerSetSize,
				     Vector definedLabels,
				     boolean noLabelBreakIsValid,
				     boolean noLabelContinueIsValid,
				     boolean inProcedure)
  {
    Translator translator = new Translator();
    translator._context.noLabelBreakIsValid = noLabelBreakIsValid;
    translator._context.noLabelContinueIsValid = noLabelContinueIsValid;
    translator._context.validLabels = definedLabels;
    translator.returnIsValid = inProcedure;
    translator.nextSlot = registerLowLimit;
    translator.registerSetSize = registerSetSize;

    try {
      translator.successor = successor;
      node.accept(translator);
    } catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException(e.getMessage());
    }

    Executable result = new Executable();
    result.registerSetSize = translator.registerSetSize;
    result.entryPoint = translator.head;
    return result;
  }

  /**
   * ɽ빽ʸ, μɾ֥ɤ.
   * 
   * @param node	Ѵץιʸ
   * @param successor	줿ɤ¹Ԥλܤ
   * 			. nullξ, 󥿥ץ꥿λ.
   * @param resultSlot	ɾ̤Ǽ٤쥸
   * @param registerLowLimit	줿ɤѹƤ褤쥸β
   * @param registerSetSize	ߤΥ쥸礭
   * @return		줿ɤΥȥݥȤɬפʥ쥸ο
   */
  public static Executable translateExpression(Node expression, Code successor,
					       int resultSlot,
					       int registerLowLimit,
					       int registerSetSize)
  {
    Translator translator = new Translator();
    translator.nextSlot = registerLowLimit;
    translator.registerSetSize = registerSetSize;

    try {
      translator.value = resultSlot;
      translator.successor = successor;
      expression.accept(translator);
    } catch (Exception e) {
      e.printStackTrace();
      throw new RuntimeException(e.getMessage());
    }

    Executable result = new Executable();
    result.registerSetSize = translator.registerSetSize;
    result.entryPoint = translator.head;
    return result;
  }

  public static Value translateLiteral(Object x) {
    if (x == ORG.ebiss.sara.parser.Literal.NULL) return Value.NULL;
    if (x instanceof Integer) return new IntegerValue(((Integer)x).intValue());
    if (x instanceof String) return new StringValue((String)x);
    if (x instanceof Double) return new DoubleValue(((Double)x).doubleValue());
    if (x instanceof Symbol) return SymbolValue.intern(((Symbol)x).name());
    if (x instanceof Boolean) return ((Boolean)x).booleanValue()
				? BooleanValue.TRUE : BooleanValue.FALSE;
    throw new RuntimeException();
  }

  protected void duplicateDefault(Node x) {
    error(x, "Duplicate default targets");
  }

  protected void duplicateLabel(Node x, String label) {
    error(x, "Duplicate label \"" + label + "\"");
  }

  protected void misplacedBreak(Node x) {
    error(x, "No enclosing while, do, for, or switch statement");
  }

  protected void misplacedContinue(Node x) {
    error(x, "No enclosing while, do, or for statement");
  }

  protected void misplacedReturn(Node x) {
    error(x, "No enclosing procedure definition");
  }

  protected void undefinedLabel(Node x, String label) {
    error(x, "Undefined label \"" + label + "\"");
  }

  protected void error(Node x, String description) {
    String	s;

    if (x == null) {
      s = "<unknown file>:0:0:";
    } else {
      String fileName = x.fileName();
      s = (fileName == null ? "<unknown file>" : fileName) + ":" +
	x.beginLine() + ":" +
	x.beginColumn() + ":";
    }

    System.out.println(s);
  }

  /*
   * ˡ˴ؤ륳:
   *
   * NodeTranslatorδ֤Υ󥿥饯VisitorѥǤ롥
   * VisitorѥŬѤ뤳Ȥˤä, if (.. instanceof ...)  
   * 򤱤뤳ȤǤ롥Node饹κѲǽ
   * 뤿ˤ, visit᥽åɤͿͤη򸷤¤
   * ɬפ롥ߤΤȤ, visit᥽åɤΰNodeΤߤǤꡤ
   * ͤϤʤʤĤޤꡤǾ٤˻ߤƤˡǡ
   * ͤѤȤưʲλĤΥ󥹥ѿTranslator
   * 
   *
   * 	successor	()
   *	value		()
   *	head		(ͤ)
   */

  private Context _context = new Context();

  protected LHSTranslator lhsTranslator = new LHSTranslator();

  /**
   * <code>return</code>ʸƤ褤ʸ̮Ǥ뤫ɤ򼨤.
   * <code>true</code>λ, <code>return</code>ʸƤ褤
   */
  protected boolean returnIsValid;

  /**
   * Υɤμ¹Ԥɬפʲۥ쥸οݻ.
   */
  protected int registerSetSize;

  /**
   * ˳Ƥ쥸ֹݻ.
   */
  protected int	nextSlot;

  /**
   * ۥ쥸դ.
   * 
   * @return		ۥ쥸ֹ
   */
  protected int getSlot() {
    int result = nextSlot++;
    if (nextSlot > registerSetSize) registerSetSize = nextSlot;
    return result;
  }

  /**
   * ٥ΤʤbreakʸиƤ褤ʸ̮Ǥ뤫ɤȽꤹ.
   * 
   * @return		trueϽиƤ褤Ȥ򼨤
   */
  protected boolean noLabelBreakIsValid() {
    return _context.noLabelBreakIsValid;
  }

  /**
   * ٥ΤʤcontinueʸиƤ褤ʸ̮Ǥ뤫ɤȽꤹ.
   * 
   * @return		trueϽиƤ褤Ȥ򼨤
   */
  protected boolean noLabelContinueIsValid() {
    return _context.noLabelContinueIsValid;
  }

  /**
   * ٥뤬Ƥ뤫ɤ򸡺.
   * @param label	٥.
   * @return		trueƤ뤳Ȥ򼨤
   */
  protected boolean isDefinedLabel(String label) {
    return _context.isDefinedLabel(label);
  }

  /**
   * ʸ̮˿Ȥ<code>Translator</code>᥽å
   * Ǥ.  ʸ̮ˤƥ٥Τʤ
   * <code>break</code>/<code>continue</code>뤫ɤ, 
   * ٥뤬줿ʤɤ˱ŬڤʽԤ.
   * <p>
   * ٥򸡽ФȤ<code>error</code>᥽åɤѤ,
   * 顼𤹤롥
   *
   * @param breakOK	trueǤС٥ΤʤbreakʸƤ褤
   *			ʸ̮˿Ȥ򼨤
   * @param continueOK	trueǤС٥ΤʤcontinueʸƤ褤
   *			ʸ̮˿Ȥ򼨤
   * @param labels	˿ʸ̮ˤ٥.
   * 			StringǤȤ٥Ǥ롥
   */
  protected void pushContext(boolean breakOK, boolean continueOK,
			     Vector labels, Node currentNode)
  {
    Context newContext = new Context();
    newContext.noLabelBreakIsValid =
      _context.noLabelBreakIsValid || breakOK;
    newContext.noLabelContinueIsValid =
      _context.noLabelContinueIsValid || continueOK;
    newContext.validLabels = labels;
    newContext.parent = _context;

    if (labels != null) {
      for (Enumeration e = labels.elements(); e.hasMoreElements(); ) {
	String label = (String)e.nextElement();
	if (isDefinedLabel(label)) {
	  duplicateLabel(currentNode, label);
	}
      }
    }

    _context = newContext;
  }

  /**
   * <code>pushContext</code>ʸ̮æФȤ
   * <code>Translator</code>.
   */
  protected void popContext() {
    _context = _context.parent;
  }

  /**
   * 륳ɤ³Ƽ¹Ԥ٤ɤ.
   * <code>visit(Node)</code>θƤӽФΰѤȤѤ.
   */
  protected Code successor;	// argument

  /**
   * 
   * <code>visit(Node)</code>θƤӽФΰѤȤѤ.
   */
  protected int inputValue;	// argument

  /**
   * ɾ̤Ǽ٤쥸ֹݻѿǤ.
   * <code>visit(Node)</code>θƤӽФ, ä˲餫ͤ׻
   * <code>Node</code>˴ؤƤӽФΰѤȤѤ
   * .
   */
  protected int value;		// argument

  /**
   * 줿ɤΥȥݥȤꤵ.
   * <code>visit(Node)</code>ͤѿǤ.
   */
  protected Code head;		// return value

  /**
   * ѥ<code>Pattern</code>֥ȤѴ.
   * 
   * @param pattern	Ѵѥιʸ
   * @return		ѥ󥪥֥
   * @see		ORG.ebiss.sara.interpreter.Pattern
   */
  public Pattern translatePattern(Node pattern) {
    PatternTranslator translator = new PatternTranslator();
    return translator.translate(pattern);
  }

  /*========================================================================
   *	ƥΡɥ饹˴ؤvisit᥽å
   *
   *   LeftHandSideѴLHSTranslatorǹԤ뤿
   * ᡤΥ饹ˤϵҤƤʤref) visit(AssignmentExpression x)
   *
   *   ѥѴ, PatternTranslator饹ǹԤ뤿,
   * Υ饹ǤƤʤ.
   *
   *========================================================================*/

  public void visit(ORG.ebiss.sara.parser.Program x) throws Exception {
    head = successor;
    for (int i = x.statements.size() - 1; i >= 0; --i) {
      Node statement = (Node)x.statements.elementAt(i);
      statement.accept(this);
      successor = head;
    }
  }

  public void visit(ORG.ebiss.sara.parser.Block x) throws Exception {
    head = successor;
    for (int i = x.statements.size() - 1; i >= 0; --i) {
      Node statement = (Node)x.statements.elementAt(i);
      statement.accept(this);
      successor = head;
    }
  }

  public void visit(ORG.ebiss.sara.parser.LabeledStatement x)
       throws Exception {
    pushContext(false, false, x.labels, x);

    Code next = successor;
    PushLabeledTargets enter = new PushLabeledTargets(x, x.labels);
    PopTarget leave = new PopTarget(x);

    successor = leave;
    x.statement.accept(this);
    Code statement = head;

    enter.breakTarget = leave;
    enter.continueTarget = statement;
    enter.successor = statement;

    leave.successor = next;

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.EmptyStatement x)
       throws Exception {
    head = successor;
  }

  public void visit(ORG.ebiss.sara.parser.StatementExpression x)
       throws Exception {
    value = getSlot();
    x.expression.accept(this);
  }

  public void visit(ORG.ebiss.sara.parser.SwitchStatement x)
       throws Exception {
    pushContext(true, false, null, x);

    Code next = successor;
    PushBreakTarget enter = new PushBreakTarget(x);
    PopTarget leave = new PopTarget(x);
    SwitchBranch branch = new SwitchBranch(x);

    Code targets[] = new Code[x.groups.size()];
    Code defaultTarget = null;
    int nlabels = 0;
    successor = leave;
    for (int i = x.groups.size() - 1; i >= 0; --i) {
      SwitchBlockStatements group =
	(SwitchBlockStatements)x.groups.elementAt(i);
      head = successor;
      for (int j = group.statements.size() - 1; j >= 0; --j) {
	((Node)group.statements.elementAt(j)).accept(this);
	successor = head;
      }
      targets[i] = head;
      if (group.hasDefault) {
	if (defaultTarget != null) {
	  duplicateDefault(x);
	}
	defaultTarget = head;
      }
      nlabels += group.labels.size();
    }

    successor = branch;
    int determinantSlot = value = getSlot();
    x.determinant.accept(this);
    Code determinant = head;

    enter.breakTarget = leave;
    enter.successor = determinant;

    branch.labels = new Object[nlabels];
    branch.targets = new Code[nlabels];
    branch.defaultTarget = defaultTarget == null ? leave : defaultTarget;
    branch.determinant = determinantSlot;
    int i = nlabels - 1;
    for (int j = x.groups.size() - 1; j >= 0; --j) {
      SwitchBlockStatements group =
	(SwitchBlockStatements)x.groups.elementAt(j);
      for (int k = group.labels.size() - 1; k >= 0; --k) {
	branch.labels[i] =
	  ((ORG.ebiss.sara.parser.Literal)group.labels.elementAt(k)).value;
	branch.targets[i] = targets[j];
	--i;
      }
    }

    leave.successor = next;

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.MatchStatement x)
       throws Exception {
    pushContext(true, false, null, x);

    Code next = successor;
    PushBreakTarget enter = new PushBreakTarget(x);
    PopTarget leave = new PopTarget(x);
    int determinantSlot = getSlot();

    head = successor = leave;
    for (int i = x.rules.size() - 1; i >= 0; --i) {
      inputValue = determinantSlot;
      ((Node)x.rules.elementAt(i)).accept(this);
      successor = head;
    }
    Code rules = head;

    successor = rules;
    value = determinantSlot;
    x.determinant.accept(this);
    Code determinant = head;

    enter.breakTarget = leave;
    enter.successor = determinant;

    leave.successor = next;

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.MatchingRule x)
       throws Exception {
    int determinantSlot = inputValue;
    Code next = successor;

    PopNameSpace pop = new PopNameSpace(x);
    pop.successor = next;

    Break end = new Break(x);

    successor = end;
    x.body.accept(this);
    Code body = head;

    if (x.condition != null) {
      ConditionalBranch branch = new ConditionalBranch(x);
      branch.condition = getSlot();
      branch.thenCode = body;
      branch.elseCode = pop;

      successor = branch;
      value = branch.condition;
      x.condition.accept(this);
    } else {
      head = body;
    }

    Match match = new Match(x);
    match.value = determinantSlot;
    match.pattern = translatePattern(x.pattern);
    match.succeeded = head;
    match.failed = pop;

    PushNameSpace enter = new PushNameSpace(x);
    enter.successor = match;

    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.IfStatement x) throws Exception {
    Code next = successor;

    ConditionalBranch branch = new ConditionalBranch(x);

    successor = branch;
    int conditionSlot = value = getSlot();
    x.condition.accept(this);
    Code condition = head;

    successor = next;
    x.thenStatement.accept(this);
    Code thenCode = head;

    Code elseCode;
    if (x.elseStatement != null) {
      successor = next;
      x.elseStatement.accept(this);
      elseCode = head;
    } else {
      elseCode = next;
    }

    branch.node = x;
    branch.condition = conditionSlot;
    branch.thenCode = thenCode;
    branch.elseCode = elseCode;

    head = condition;
  }

  public void visit(ORG.ebiss.sara.parser.WhileStatement x)
       throws Exception {
    pushContext(true, true, null, x);

    Code next = successor;

    PushBreakContinueTargets enter = new PushBreakContinueTargets(x);
    PopTarget leave = new PopTarget(x);
    ConditionalBranch branch = new ConditionalBranch(x);

    successor = branch;
    int conditionSlot = value = getSlot();
    x.condition.accept(this);
    Code condition = head;

    successor = condition;
    x.body.accept(this);
    Code body = head;

    enter.breakTarget = leave;
    enter.continueTarget = condition;
    enter.successor = condition;

    branch.condition = conditionSlot;
    branch.thenCode = body;
    branch.elseCode = leave;

    leave.successor = next;

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.DoStatement x)
       throws Exception {
    pushContext(true, true, null, x);

    Code next = successor;

    PopTarget leave = new PopTarget(x);
    leave.successor = next;

    ConditionalBranch branch = new ConditionalBranch(x);
    branch.condition = getSlot();
    branch.elseCode = leave;

    successor = branch;
    value = branch.condition;
    x.condition.accept(this);
    Code condition = head;

    successor = condition;
    x.body.accept(this);
    Code body = head;
    branch.thenCode = body;

    PushBreakContinueTargets enter = new PushBreakContinueTargets(x);
    enter.breakTarget = leave;
    enter.continueTarget = body;
    enter.successor = body;

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.ForStatement x)
       throws Exception {
    pushContext(true, true, null, x);

    Code next = successor;

    PushBreakContinueTargets enter = new PushBreakContinueTargets(x);
    PopTarget leave = new PopTarget(x);
    Code branch = null;
    Code condition = null;
    int conditionSlot = 0;

    if (x.condition != null) {
      successor = branch = new ConditionalBranch(x);
      conditionSlot = value = getSlot();
      x.condition.accept(this);
      condition = head;
      successor = head;
    } else {
      successor = branch = new Branch(x);
    }

    head = successor;
    for (int i = x.updater.size() - 1; i >= 0; --i) {
      ((Node)x.updater.elementAt(i)).accept(this);
      successor = head;
    }
    Code updater = head;

    successor = updater;
    x.body.accept(this);
    Code body = head;

    if (x.condition != null) {
      successor = condition;
    } else {
      successor = body;
    }
    head = successor;
    for (int i = x.initializer.size() - 1; i >= 0; --i) {
      ((Node)x.initializer.elementAt(i)).accept(this);
      successor = head;
    }
    Code initializer = head;

    enter.breakTarget = leave;
    enter.continueTarget = updater;
    enter.successor = initializer;

    leave.successor = next;

    if (x.condition != null) {
      ConditionalBranch cbranch = (ConditionalBranch)branch;
      cbranch.condition = conditionSlot;
      cbranch.thenCode = body;
      cbranch.elseCode = leave;
    } else {
      ((Branch)branch).target = body;
    }

    popContext();
    head = enter;
  }

  public void visit(ORG.ebiss.sara.parser.BreakStatement x)
       throws Exception {
    if (x.label == null) {
      if (! noLabelBreakIsValid()) {
	misplacedBreak(x);
	return;
      }
      head = new Break(x);
    } else {
      if (! isDefinedLabel(x.label)) {
	undefinedLabel(x, x.label);
	return;
      }
      BreakFrom result = new BreakFrom(x);
      result.label = x.label;
      head = result;
    }
  }

  public void visit(ORG.ebiss.sara.parser.ContinueStatement x)
       throws Exception {
    if (x.label == null) {
      if (! noLabelContinueIsValid()) {
	misplacedContinue(x);
	return;
      }
      head = new Continue(x);
    } else {
      if (! isDefinedLabel(x.label)) {
	undefinedLabel(x, x.label);
	return;
      }
      ContinueTo result = new ContinueTo(x);
      result.label = x.label;
      head = result;
    }
  }

  public void visit(ORG.ebiss.sara.parser.ReturnStatement x)
       throws Exception {
    if (! returnIsValid) {
      misplacedReturn(x);
      return;
    }

    Return wayout = new Return(x);
    wayout.value = getSlot();

    successor = wayout;
    value = wayout.value;
    if (x.value != null) {
      x.value.accept(this);
      return;
    }
    LoadLiteral literal = new LoadLiteral(x, Value.UNSPECIFIED);
    literal.slot = wayout.value;
    literal.successor = wayout;
    head = literal;
  }

  public void visit(ORG.ebiss.sara.parser.AssertStatement x)
       throws Exception
  {
    Code next = successor;
    Assert assert = new Assert(x);

    successor = assert;
    int conditionSlot = value = getSlot();
    x.condition.accept(this);
    Code condition = head;

    assert.condition = conditionSlot;
    assert.successor = next;

    head = condition;
  }

  public void visit(ORG.ebiss.sara.parser.UseStatement x)
       throws Exception {
    head = successor;
    for (int i = x.body.size() - 1; i >= 0; --i) {
      ((Node)x.body.elementAt(i)).accept(this);
      successor = head;
    }
  }

  public void visit(ORG.ebiss.sara.parser.BinaryOpExpression x)
       throws Exception {
    Code next = successor;
    int binopSlot = value;

    BinaryOp binop = new BinaryOp(x);

    successor = binop;
    int rightSlot = value = getSlot();
    x.right.accept(this);
    Code right = head;

    successor = right;
    int leftSlot = value = getSlot();
    x.left.accept(this);
    Code left = head;

    binop.opcode = x.operator;
    binop.left = leftSlot;
    binop.right = rightSlot;
    binop.slot = binopSlot;
    binop.successor = next;

    head = left;
  }

  public void visit(ORG.ebiss.sara.parser.ConditionalOpExpression x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    ConditionalBranch branch1 = new ConditionalBranch(x);
    ConditionalBranch branch2 = new ConditionalBranch(x);
    LoadLiteral loadTrue = new LoadLiteral(x, BooleanValue.TRUE);
    LoadLiteral loadFalse = new LoadLiteral(x, BooleanValue.FALSE);

    successor = branch1;
    int leftSlot = value = getSlot();
    x.left.accept(this);
    Code left = head;

    successor = branch2;
    int rightSlot = value = getSlot();
    x.right.accept(this);
    Code right = head;

    branch1.condition = leftSlot;
    branch2.condition = rightSlot;
    switch (x.operator) {
    case ConditionalOpExpression.CONDITIONAL_AND:
      branch1.thenCode = right;
      branch1.elseCode = loadFalse;
      break;
    case ConditionalOpExpression.CONDITIONAL_OR:
      branch1.thenCode = loadTrue;
      branch1.elseCode = right;
      break;
    default:
      throw new RuntimeException();
    }
    branch2.thenCode = loadTrue;
    branch2.elseCode = loadFalse;

    loadTrue.slot = loadFalse.slot = resultSlot;
    loadTrue.successor = loadFalse.successor = next;

    head = left;
  }

  public void visit(ORG.ebiss.sara.parser.ConditionalExpression x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    ConditionalBranch branch = new ConditionalBranch(x);

    successor = branch;
    int conditionSlot = value = getSlot();
    x.condition.accept(this);
    Code condition = head;

    successor = next;
    value = resultSlot;
    x.thenExpression.accept(this);
    Code thenCode = head;

    successor = next;
    value = resultSlot;
    x.elseExpression.accept(this);
    Code elseCode = head;

    branch.condition = conditionSlot;
    branch.thenCode = thenCode;
    branch.elseCode = elseCode;

    head = condition;
  }

  public void visit(ORG.ebiss.sara.parser.UnaryOpExpression x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    UnaryOp uop = new UnaryOp(x);

    successor = uop;
    int operandSlot = value = getSlot();
    x.expression.accept(this);
    Code operand = head;

    uop.opcode = x.operator;
    uop.operand = operandSlot;
    uop.slot = resultSlot;
    uop.successor = next;

    head = operand;
  }

  public void visit(ORG.ebiss.sara.parser.AssignmentExpression x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    inputValue = resultSlot;
    successor = next;
    x.left.accept(lhsTranslator);
    Code store = head;

    if (x.operator == ORG.ebiss.sara.parser.AssignmentExpression.ASSIGN) {
      successor = store;
      value = resultSlot;
      x.right.accept(this);
    } else {
      BinaryOp op = new BinaryOp(x);

      successor = op;
      int rightSlot = value = getSlot();
      x.right.accept(this);
      Code right = head;

      successor = right;
      int leftSlot = value = getSlot();
      x.left.accept(this);
      Code left = head;

      op.opcode = x.operator;
      op.slot = resultSlot;
      op.right = rightSlot;
      op.left = leftSlot;
      op.successor = store;

      head = left;
    }
  }

  public void visit(ORG.ebiss.sara.parser.IncOrDecExpression x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    int preinc = 0;
    int postinc = 0;
    Node current;
    for (current = x;
	 current instanceof IncOrDecExpression;
	 current = ((IncOrDecExpression)current).expression) {
      switch (((IncOrDecExpression)current).operator) {
      case IncOrDecExpression.PREINCREMENT:
	++preinc;
	break;
      case IncOrDecExpression.PREDECREMENT:
	--preinc;
	break;
      case IncOrDecExpression.POSTINCREMENT:
	++postinc;
	break;
      case IncOrDecExpression.POSTDECREMENT:
	--postinc;
	break;
      default:
	throw new RuntimeException();
      }
    }

    if (postinc != 0) {
      int assignSlot = inputValue = getSlot();
      current.accept(lhsTranslator);
      BinaryOp binop = new BinaryOp(x);
      binop.left = resultSlot;
      binop.right = getSlot();
      binop.opcode = ORG.ebiss.sara.parser.BinaryOpExpression.PLUS;
      binop.slot = assignSlot;
      binop.successor = head;
      LoadLiteral literal = new LoadLiteral(x, new IntegerValue(postinc));
      literal.successor = binop;
      literal.slot = binop.right;
      successor = literal;
    }

    if (preinc == 0) {
      value = resultSlot;
      current.accept(this);
    } else {
      inputValue = resultSlot;
      current.accept(lhsTranslator);
      BinaryOp binop = new BinaryOp(x);
      binop.left = getSlot();
      binop.right = getSlot();
      binop.opcode = ORG.ebiss.sara.parser.BinaryOpExpression.PLUS;
      binop.slot = resultSlot;
      binop.successor = head;
      LoadLiteral literal = new LoadLiteral(x, new IntegerValue(preinc));
      literal.slot = binop.right;
      literal.successor = binop;
      successor = literal;
      value = binop.left;
      current.accept(this);
    }
  }

  public void visit(ORG.ebiss.sara.parser.VariableReference x)
       throws Exception {
    ReadVariable read = new ReadVariable(x);
    read.name = x.name;
    read.slot = value;
    read.successor = successor;
    head = read;
  }

  public void visit(ORG.ebiss.sara.parser.ParenthesizedExpression x)
       throws Exception {
    x.expression.accept(this);
  }

  public void visit(ORG.ebiss.sara.parser.Literal x) throws Exception {
    LoadLiteral literal = new LoadLiteral(x, translateLiteral(x.value));
    literal.slot = value;
    literal.successor = successor;
    head = literal;
  }

  public void visit(ORG.ebiss.sara.parser.ThisExpression x)
       throws Exception {
    ReadThis read = new ReadThis(x);
    read.slot = value;
    read.successor = successor;
    head = read;
  }

  public void visit(ORG.ebiss.sara.parser.SequenceConstruction x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    ConstructList	list = null;
    ConstructVector	vector = null;

    switch (x.type) {
    case ORG.ebiss.sara.parser.SequenceConstruction.LIST:
      list = new ConstructList(x);
      successor = list;
      break;

    case ORG.ebiss.sara.parser.SequenceConstruction.VECTOR:
      vector = new ConstructVector(x);
      successor = vector;
      break;

    default:
      throw new RuntimeException();
    }

    int[] elementSlots = new int[x.elements.size()];
    head = successor;
    for (int i = elementSlots.length - 1; i >= 0; --i) {
      elementSlots[i] = value = getSlot();
      ((Node)x.elements.elementAt(i)).accept(this);
      successor = head;
    }
    Code elementCode = head;

    switch (x.type) {
    case ORG.ebiss.sara.parser.SequenceConstruction.LIST:
      list.elements = elementSlots;
      list.slot = resultSlot;
      list.successor = next;
      break;

    case ORG.ebiss.sara.parser.SequenceConstruction.VECTOR:
      vector.elements = elementSlots;
      vector.slot = resultSlot;
      vector.successor = next;
      break;

    default:
      throw new RuntimeException();
    }

    head = elementCode;
  }

  public void visit(ORG.ebiss.sara.parser.ClosureConstruction x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    if (x.procedures.size() <= 0) throw new RuntimeException();

    ConstructClosure cons = new ConstructClosure(x);

    Executable[] procedures = new Executable[x.procedures.size()];
    Pattern[][] patterns = new Pattern[x.procedures.size()][];
    for (int i = x.procedures.size() - 1; i >= 0; --i) {
      ProcedureConstruction proc =
	(ProcedureConstruction)x.procedures.elementAt(i);
      patterns[i] = new Pattern[proc.arguments.size()];
      for (int j = proc.arguments.size() - 1; j >= 0; --j) {
	patterns[i][j] = translatePattern((Node)proc.arguments.elementAt(j));
      }
      procedures[i] =
	translate(proc, successor, 0, 0, null, false, false, true);
    }

    cons.procedures = procedures;
    cons.patterns = patterns;
    cons.slot = resultSlot;
    cons.successor = next;

    head = cons;
  }

  public void visit(ORG.ebiss.sara.parser.ProcedureConstruction x)
       throws Exception {
    Code next = successor;

    LoadLiteral literal = new LoadLiteral(x, Value.UNSPECIFIED);
    ConditionalBranch branch = new ConditionalBranch(x);
    Return end = new Return(x);

    successor = literal;
    x.body.accept(this);
    Code body = head;

    Code condition = null;
    if (x.condition != null) {
      successor = branch;
      int conditionSlot = value = getSlot();
      x.condition.accept(this);
      condition = head;

      branch.condition = conditionSlot;
      branch.thenCode = body;
      branch.elseCode = null;
    }

    literal.slot = getSlot();
    literal.successor = end;

    end.value = literal.slot;

    head = condition == null ? body : condition;
  }

  public void visit(ORG.ebiss.sara.parser.ObjectConstruction x)
       throws Exception {
    Code next = successor;
    int resultSlot = value;

    PushReadOnlyNameSpace push = new PushReadOnlyNameSpace(x);
    PopNameSpace pop = new PopNameSpace(x);
    ConstructObject object = new ConstructObject(x);

    pop.successor = next;

    head = successor = pop;
    for (int i = x.members.size() - 1; i >= 0; --i) {
      inputValue = resultSlot;
      ((Node)x.members.elementAt(i)).accept(this);
      successor = head;
    }

    push.successor = head;
    push.object = resultSlot;

    object.slot = resultSlot;
    object.successor = push;

    head = object;
  }

  public void visit(ORG.ebiss.sara.parser.MethodDefinition x)
       throws Exception {

    Code next = successor;
    int objectSlot = inputValue;

    DefineMethod define = new DefineMethod(x);

    if (x.procedures.size() <= 0) throw new RuntimeException();

    UserDefinedMethod[] methods = new UserDefinedMethod[x.procedures.size()];

    for (int i = x.procedures.size() - 1; i >= 0; --i) {
      ProcedureConstruction proc =
	(ProcedureConstruction)x.procedures.elementAt(i);
      Pattern[] patterns = new Pattern[proc.arguments.size()];
      for (int j = proc.arguments.size() - 1; j >= 0; --j) {
	patterns[j] = translatePattern((Node)proc.arguments.elementAt(j));
      }
      Executable code = translate(proc, successor, 0, 0, null,
				  false, false, true);
      methods[i] = new UserDefinedMethod(patterns, code);
    }

    String[] modifiers = new String[x.modifiers.size()];
    for (int i = 0; i < modifiers.length; ++i) {
      modifiers[i] = (String)x.modifiers.elementAt(i);
    }

    define.name = x.name;
    define.modifiers = modifiers;
    define.methods = methods;
    define.object = objectSlot;
    define.successor = next;

    head = define;
  }

  public void visit(ORG.ebiss.sara.parser.InstanceVariableDef x)
       throws Exception {
    Code next = successor;
    int objectSlot = inputValue;

    DefineField define = new DefineField(x);

    successor = define;
    int resultSlot = value = getSlot();
    x.value.accept(this);
    Code initializer = head;

    String[] modifiers = new String[x.modifiers.size()];
    for (int i = 0; i < modifiers.length; ++i) {
      modifiers[i] = (String)x.modifiers.elementAt(i);
    }

    define.name = x.name;
    define.modifiers = modifiers;
    define.value = resultSlot;
    define.object = objectSlot;
    define.successor = next;

    head = initializer;
  }

  public void visit(ORG.ebiss.sara.parser.FieldAccess x) throws Exception {
    Code next = successor;
    int result = value;

    ReadField read = new ReadField(x);

    successor = read;
    int objectSlot = value = getSlot();
    x.object.accept(this);
    Code object = head;

    read.successor = next;
    read.slot = result;
    read.object = objectSlot;
    read.field = x.field;

    head = object;
  }

  public void visit(ORG.ebiss.sara.parser.SuperFieldAccess x)
       throws Exception {
    ReadSuperField read = new ReadSuperField(x);
    read.successor = successor;
    read.slot = value;
    read.field = x.field;

    head = read;
  }

  public void visit(ORG.ebiss.sara.parser.VectorAccess x)
       throws Exception {
    Code next = successor;
    int result = value;

    ReadField read = new ReadField(x);
    Call call = new Call(x);

    successor = read;
    int indexSlot = value = getSlot();
    x.index.accept(this);
    Code index = head;

    successor = index;
    int objectSlot = value = getSlot();
    x.vector.accept(this);
    Code object = head;

    read.field = "at";
    read.object = objectSlot;
    read.successor = call;
    read.slot = getSlot();

    call.applicant = read.slot;
    call.arguments = new int[1];
    call.arguments[0] = indexSlot;
    call.slot = result;
    call.successor = next;

    head = object;
  }

  public void visit(ORG.ebiss.sara.parser.Invocation x) throws Exception {
    Code next = successor;
    int result = value;

    Call call = new Call(x);

    int[] arguments = new int[x.arguments.size()];
    head = successor = call;
    for (int i = x.arguments.size() - 1; i >= 0; --i) {
      arguments[i] = value = getSlot();
      ((Node)x.arguments.elementAt(i)).accept(this);
      successor = head;
    }
    Code argumentsCode = head;

    successor = argumentsCode;
    int applicant = value = getSlot();
    x.applicant.accept(this);
    Code applicantCode = head;

    call.applicant = applicant;
    call.arguments = arguments;
    call.slot = result;
    call.successor = next;

    head = applicantCode;
  }

  public void visit(ORG.ebiss.sara.parser.CallCCExpression x)
       throws Exception
  {
    Translator translator = new Translator();
    translator._context.noLabelBreakIsValid = false;
    translator._context.noLabelContinueIsValid = false;
    translator._context.validLabels = null;
    translator.returnIsValid = true;
    translator.nextSlot = 0;
    translator.registerSetSize = 0;
    translator.successor = null;

    Return end = new Return(x);
    end.value = translator.getSlot();

    LoadLiteral literal = new LoadLiteral(x, Value.UNSPECIFIED);
    literal.slot = end.value;
    literal.successor = end;

    translator.successor = literal;
    x.body.accept(translator);

    Executable body = new Executable();
    body.entryPoint = translator.head;
    body.registerSetSize = translator.registerSetSize;

    CallCC callcc = new CallCC(x);
    callcc.slot = value;
    callcc.successor = successor;
    callcc.identifier = x.identifier;
    callcc.body = body;

    head = callcc;
  }

  class LHSTranslator extends ORG.ebiss.sara.parser.PartialNodeVisitor {

    public void visit(ORG.ebiss.sara.parser.BinaryOpExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ConditionalOpExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ConditionalExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.AssignmentExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.UnaryOpExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void
    visit(ORG.ebiss.sara.parser.IncOrDecExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.Literal x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ThisExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.SequenceConstruction x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ClosureConstruction x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ProcedureConstruction x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.ObjectConstruction x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.MethodDefinition x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.InstanceVariableDef x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.SuperFieldAccess x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.Invocation x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.CallCCExpression x)
	 throws Exception
    {
      Translator.this.visit(x);
    }

    public void visit(ORG.ebiss.sara.parser.VariableReference x)
	 throws Exception
    {
      WriteVariable write = new WriteVariable(x);
      write.name = x.name;
      write.value = inputValue;
      write.successor = successor;
      head = write;
    }

    public void visit(ORG.ebiss.sara.parser.VariableDefinition x)
	 throws Exception {
      DefineVariable def = new DefineVariable(x);
      def.name = x.name;
      def.value = inputValue;
      def.successor = successor;
      head = def;
    }

    public void visit(ORG.ebiss.sara.parser.FieldAccess x)
	 throws Exception
    {
      Code next = successor;
      WriteField field = new WriteField(x);
      field.value = inputValue;

      successor = field;
      int objectSlot = value = getSlot();
      x.object.accept(Translator.this);
      Code object = head;

      field.object = objectSlot;
      field.fieldName = x.field;
      field.successor = next;

      head = object;
    }

    public void visit(ORG.ebiss.sara.parser.VectorAccess x)
	 throws Exception
    {
      Code next = successor;
      int operandSlot = inputValue;

      ReadField read = new ReadField(x);
      Call call = new Call(x);

      successor = read;
      int indexSlot = value = getSlot();
      x.index.accept(Translator.this);
      Code index = head;

      successor = index;
      int objectSlot = value = getSlot();
      x.vector.accept(Translator.this);
      Code object = head;

      read.field = "put";
      read.object = objectSlot;
      read.successor = call;
      read.slot = getSlot();

      call.applicant = read.slot;
      call.arguments = new int[2];
      call.arguments[0] = indexSlot;
      call.arguments[1] = operandSlot;
      call.slot = getSlot();
      call.successor = next;

      head = object;
    }

    public void visit(ORG.ebiss.sara.parser.ParenthesizedExpression x)
	 throws Exception
    {
      x.expression.accept(this);
    }
  }

  static class Context {
    boolean	noLabelBreakIsValid;
    boolean	noLabelContinueIsValid;
    Vector	validLabels;
    Context	parent;

    public boolean isDefinedLabel(String label) {
      if (validLabels != null && validLabels.contains(label)) return true;
      if (parent == null) return false;
      return parent.isDefinedLabel(label);
    }
  }
}

class UserDefinedMethod implements Method {
  protected Pattern[]		_patterns;
  protected Executable		_code;
  public Scope			environment;

  public UserDefinedMethod(Pattern[] patterns, Executable code) {
    _patterns = patterns;
    _code = code;
  }

  public Code apply(Interpreter interpreter,
		    SpdValue container, SpdValue receiver) {
    if (interpreter.arity() != _patterns.length) return null;
    interpreter.setScope(environment);
    interpreter.pushReadOnlyNameSpace(receiver);
    LocalNameSpace space = new LocalNameSpace();
    interpreter.pushNameSpace(space);
    space.define(".container", container);
    for (int i = 0; i < _patterns.length; ++i) {
      Value v = interpreter.readArgument(i);
      if (! _patterns[i].match(interpreter, v)) {
	interpreter.resetNameSpace();
	return null;
      }
    }
    interpreter.createRegisterSet(_code.registerSetSize);
    return _code.entryPoint;
  }

  public Object getPrintableRepresentation(PrintingContext context) {
    StringBuffer buffer = new StringBuffer("@");
    buffer.append(_code.entryPoint.hashCode());
    buffer.append("(");
    if (_patterns.length > 0) {
      buffer.append("*");
      for (int i = 1; i < _patterns.length; ++i) {
	buffer.append(",*");
      }
    }
    buffer.append(")");
    return buffer.toString();
  }

  public boolean checkArity(int arity) { return _patterns.length == arity; }
}

final class LoadLiteral extends ExpressionCode {
  public Value	value;

  LoadLiteral(Node x, Value y) {
    super(x);
    value = y;
  }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.writeRegister(slot, value);
    return successor;
  }
}

final class UnaryOp extends ExpressionCode {
  public int	operand;
  public int	opcode;

  UnaryOp(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readRegister(operand);
    try {
      return value.unaryOp(interpreter, opcode, slot, successor);
    } catch (InvalidOperationException usoe) {
      invalidOperation(opcode, value);
    } catch (CoerceFailedException cfe) {
      domainError(value);
    }
    return successor;
  }
}

final class BinaryOp extends ExpressionCode {
  public int	right;
  public int	left;
  public int	opcode;

  BinaryOp(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readRegister(left);
    try {
      return value.binaryOp(interpreter, opcode, right, slot, successor);
    } catch (InvalidOperationException usoe) {
      invalidOperation(opcode, value, interpreter.readRegister(right));
    } catch (CoerceFailedException cfe) {
      domainError(interpreter.readRegister(right));
    }
    return successor;
  }
}

final class ReadVariable extends ExpressionCode {
  public String	name;

  ReadVariable(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readVariable(name);
    if (value == null) undefinedVariable(name);
    interpreter.writeRegister(slot, value);
    return successor;
  }
}

final class ReadThis extends ExpressionCode {
  ReadThis(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readVariable("this");
    if (value == null) misplacedThis();
    interpreter.writeRegister(slot, value);
    return successor;
  }
}

final class Branch extends Code {
  public Code	target;

  Branch(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    return target;
  }
}

final class ConditionalBranch extends Code {
  public int	condition;
  public Code	thenCode;
  public Code	elseCode;

  ConditionalBranch(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readRegister(condition);
    try {
      return value.booleanValue() ? thenCode : elseCode;
    } catch (CoerceFailedException cfe) {
      domainError(value, "boolean");
    }
    return null;
  }
}

final class WriteVariable extends Code {
  public String	name;
  public int	value;
  public Code	successor;

  WriteVariable(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value v = interpreter.readRegister(value);
    try {
      if (! interpreter.writeVariable(name, v)) {
	throw new EvaluationException("Undefined variable " + name);
      }
    } catch(InvalidOperationException usoe) {
      usoe.printStackTrace();
      throw new RuntimeException(usoe.getMessage());
    }
    return successor;
  }  
}

final class DefineVariable extends Code {
  public String	name;
  public int	value;
  public Code	successor;

  DefineVariable(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value v = interpreter.readRegister(value);
    try {
      if (! interpreter.defineVariable(name, v)) {
	System.err.println("Duplicate definition of variable " + name);
	interpreter.writeVariable(name, v);
      }
    } catch(InvalidOperationException usoe) {
      usoe.printStackTrace();
      throw new RuntimeException(usoe.getMessage());
    }
    return successor;
  }  
}

final class ConstructList extends ExpressionCode {
  public int[]	elements;

  ConstructList(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    ListValue value = (ListValue)ListValue.anList.copy();
    Container container = (Container)value.getJavaValue();
    for (int i = 0; i < elements.length; ++i) {
      container.add(interpreter.readRegister(elements[i]));
    }
    interpreter.writeRegister(slot, value);
    return successor;
  }
}

final class ConstructVector extends ExpressionCode {
  public int[]	elements;

  ConstructVector(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    VectorValue value = (VectorValue)VectorValue.anVector.copy();
    Container container = (Container)value.getJavaValue();
    for (int i = 0; i < elements.length; ++i) {
      container.add(interpreter.readRegister(elements[i]));
    }
    interpreter.writeRegister(slot, value);
    return successor;
  }
}

final class ConstructClosure extends ExpressionCode {
  public Executable[]	procedures;
  public Pattern[][]	patterns;

  ConstructClosure(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    ClosureValue result = new ClosureValue(procedures, patterns);
    result.environment = interpreter.getScope();
    interpreter.writeRegister(slot, result);
    return successor;
  }
}

final class ClosureValue extends ApplicableValue {
  public Scope	environment;
  Executable[]	procedures;
  Pattern[][]	patterns;

  ClosureValue(Executable[] procedures, Pattern[][] patterns) {
    this.procedures = procedures;
    this.patterns = patterns;
  }

  public Code apply(Interpreter interpreter, int[] arguments,
		    int resultSlot, Code successor)
  {
    Code nextProc =
      new NextProcedureInClosure(this, interpreter.currentNode());
    interpreter.pushFrame(arguments, resultSlot, successor, nextProc);
    return nextProc;
  }

  public boolean checkArity(int arity) {
    for (int i = 0; i < patterns.length; ++i) {
      if (patterns[i].length == arity) return true;
    }
    return false;
  }
}

final class NextProcedureInClosure extends Code {
  ClosureValue	closure;
  public int	index = 0;

  NextProcedureInClosure(ClosureValue x, Node y) { super(y); closure = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    LocalNameSpace ns = new LocalNameSpace();
    interpreter.setScope(closure.environment);
    interpreter.pushNameSpace(ns);
    Executable exec = null;
  nextProcedure:
    for (;;) {
      ns.clear();
      if (closure.procedures.length <= index) noMatchingProcedure();
      exec = closure.procedures[index];
      Pattern[] pats = closure.patterns[index];
      ++index;
      if (interpreter.arity() != pats.length) continue;
      for (int i = 0; i < pats.length; ++i) {
	if (! pats[i].match(interpreter, interpreter.readArgument(i))) {
	  continue nextProcedure;
	}
      }
      break;
    }
    interpreter.createRegisterSet(exec.registerSetSize);
    return exec.entryPoint;
  }
}

final class PushNameSpace extends Code {
  public Code	successor;

  PushNameSpace(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    LocalNameSpace nameSpace = new LocalNameSpace();
    interpreter.pushNameSpace(nameSpace);
    return successor;
  }
}

final class PushReadOnlyNameSpace extends Code {
  public Code	successor;
  public int	object;

  PushReadOnlyNameSpace(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.pushReadOnlyNameSpace(interpreter.readRegister(object));
    return successor;
  }
}

final class PopNameSpace extends Code {
  public Code	successor;

  PopNameSpace(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.popNameSpace();
    return successor;
  }
}

final class PushBreakTarget extends Code {
  public Code	breakTarget;
  public Code	successor;

  PushBreakTarget(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.pushBreakTarget(breakTarget);
    return successor;
  }
}

final class PushBreakContinueTargets extends Code {
  public Code	breakTarget;
  public Code	continueTarget;
  public Code	successor;

  PushBreakContinueTargets(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.pushBreakContinueTargets(breakTarget, continueTarget);
    return successor;
  }
}

final class PushLabeledTargets extends Code {
  public String[]	labels;
  public Code		breakTarget;
  public Code		continueTarget;
  public Code		successor;

  PushLabeledTargets(Node x, Vector labels) {
    node = x;
    this.labels = new String[labels.size()];
    for (int i = labels.size() - 1; i >= 0; --i) {
      this.labels[i] = (String)labels.elementAt(i);
    }
  }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.pushLabeledTargets(labels, breakTarget, continueTarget);
    return successor;
  }
}

final class PopTarget extends Code {
  public Code	successor;

  PopTarget(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    interpreter.popTarget();
    return successor;
  }
}

final class Break extends Code {
  Break(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Code successor = interpreter.doBreak();
    if (successor == null) misplacedBreak();
    return successor;
  }
}

final class BreakFrom extends Code {
  public String	label;

  BreakFrom(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Code successor = interpreter.doBreak(label);
    if (successor == null) undefinedLabel(label);
    return successor;
  }
}

final class Continue extends Code {
  Continue(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Code successor = interpreter.doContinue();
    if (successor == null) misplacedContinue();
    return successor;
  }
}

final class ContinueTo extends Code {
  public String	label;

  ContinueTo(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Code successor = interpreter.doContinue(label);
    if (successor == null) misplacedContinue();
    return successor;
  }
}

final class Return extends Code {
  public int	value;

  Return(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    return interpreter.popFrameAndReturn(interpreter.readRegister(value));
  }
}

final class Assert extends Code {
  public int	condition;
  public Code	successor;

  Assert(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value c = interpreter.readRegister(condition);
    try {
      if (! c.booleanValue()) {
	throw new BreakPointException("ASSERTION FAILURE", successor);
      }
    } catch (CoerceFailedException cfe) {
      domainError(c, "boolean");
    }
    return successor;
  }
}

final class SwitchBranch extends Code {
  public Object[]	labels;
  public Code[]		targets;
  public Code		defaultTarget;
  public int		determinant;

  SwitchBranch(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readRegister(determinant);
    for (int i = 0; i < labels.length; ++i) {
      if (value.equalsToLabel(labels[i])) return targets[i];
    }
    return defaultTarget;
  }
}

final class Match extends Code {
  public int		value;
  public Pattern	pattern;
  public Code		succeeded;
  public Code		failed;

  Match(Node x) { node = x; }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    return pattern.match(interpreter, interpreter.readRegister(value))
      ? succeeded : failed;
  }
}

final class ConstructObject extends ExpressionCode {
  ConstructObject(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    SpdValue object = new SpdValue();
    interpreter.writeRegister(slot, object);
    return successor;
  }
}

final class DefineMethod extends Code {
  public String[]		modifiers;
  public String			name;
  public UserDefinedMethod[]	methods;
  public int			object;
  public Code			successor;

  DefineMethod(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    SpdValue target = (SpdValue)interpreter.readRegister(object);
    for (int i = 0; i < methods.length; ++i) {
      methods[i].environment = interpreter.getScope();
    }
    try {
      target.defineMethod(modifiers, name, methods);
    } catch (InvalidOperationException usoe) {
      failedToDefineMethod(target, modifiers, name);
    }
    return successor;
  }
}

final class DefineField extends Code {
  public String[]	modifiers;
  public String		name;
  public int		value;
  public int		object;
  public Code		successor;

  DefineField(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    SpdValue target = (SpdValue)interpreter.readRegister(object);
    Value v = interpreter.readRegister(value);
    try {
      target.defineVariable(modifiers, name, v);
    } catch (InvalidOperationException usoe) {
      failedToDefineField(target, modifiers, name, v);
    }
    return successor;
  }
}

final class ConstructJavaClass extends ExpressionCode {
  public int	className;

  ConstructJavaClass(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value classNameValue = null;
    String name = null;
    try {
      classNameValue = interpreter.readRegister(className);
      name = classNameValue.stringValue();
      interpreter.writeRegister(slot, new JavaValue(Class.forName(name)));
    } catch (CoerceFailedException cfe) {
      domainError(classNameValue, "string");
    } catch (ClassNotFoundException cnfe) {
      javaClassNotFound(name);
    }
    return successor;
  }
}

final class ReadField extends ExpressionCode {
  public int	object;
  public String	field;

  ReadField(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value o = interpreter.readRegister(object);
    Value v = o.read(field);
    if (v == null) undefinedField(o, field);
    interpreter.writeRegister(slot, v);
    return successor;
  }
}

final class ReadSuperField extends ExpressionCode {
  public String	field;

  ReadSuperField(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    SpdValue o = (SpdValue)interpreter.readVariable(".container");
    Value v = o.readSuper(field, (SpdValue)interpreter.readVariable("this"));
    if (v == null) undefinedField(o, field);
    interpreter.writeRegister(slot, v);
    return successor;
  }
}

final class WriteField extends Code {
  public int		object;
  public String		fieldName;
  public int		value;
  public Code		successor;

  WriteField(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value o = interpreter.readRegister(object);
    Value v = interpreter.readRegister(value);
    try {
      if (! o.write(fieldName, v)) undefinedField(o, fieldName);
    } catch(InvalidOperationException usoe) {
      failedToWriteField(o, fieldName, v);
    }
    return successor;
  }
}

final class Call extends ExpressionCode {
  public int	applicant;
  public int[]	arguments;

  Call(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    Value value = interpreter.readRegister(applicant);
    if (! (value instanceof ApplicableValue)) {
      domainError(value, "applicable value");
    }
    return
      ((ApplicableValue)value).apply(interpreter, arguments, slot, successor);
  }
}

final class CallCC extends ExpressionCode {
  public String		identifier;
  public Executable	body;

  public CallCC(Node x) { super(x); }

  public Code execute(Interpreter interpreter) throws EvaluationException {
    try {
      Continuation continuation = new Continuation();
      continuation.context = interpreter.saveContext();
      continuation.slot = slot;
      continuation.successor = successor;
      Scope scope = interpreter.getScope();
      interpreter.pushFrame(new int[0], slot, successor, new CallCCNext());
      interpreter.createRegisterSet(body.registerSetSize);
      interpreter.setScope(scope);
      interpreter.pushNameSpace(new LocalNameSpace());
      interpreter.defineVariable(identifier, continuation);
      return body.entryPoint;
    } catch (InvalidOperationException uoe) {
      uoe.printStackTrace();
      throw new RuntimeException();
    }
  }

  static class CallCCNext extends Code {
    public Code execute(Interpreter interpreter) throws EvaluationException {
      noMatchingProcedure();
      return null;
    }
  }

  class Continuation extends ApplicableValue {
    public Context	context;
    public int		slot;
    public Code		successor;

    public Code apply(Interpreter interpreter, int[] arguments,
		      int resultSlot, Code returnPoint)
	 throws EvaluationException
    {
      if (arguments.length != 1) {
	throw new EvaluationException("arity of continuations is 1");
      }
      Value argument = interpreter.readRegister(arguments[0]);
      interpreter.restoreContext(context);
      interpreter.writeRegister(slot, argument);
      return successor;
    }

    public boolean checkArity(int arity) { return arity == 1; }
  }
}
