package ORG.ebiss.sara.server;

import java.io.*;
import java.net.*;
import java.util.*;

public class Client {
  public String		hostname;
  public PrintStream	err = System.err;
  protected ClientPort	_port;
  protected Hashtable	_adapters = new Hashtable();

  public Client(String hostname) throws IOException, UnknownHostException {
    this.hostname = hostname;
    _port = new ClientPort(this);
  }

  public synchronized void reset() throws IOException {
    _port.close();
    _port = new ClientPort(this);
  }

  public synchronized SimValueAdapter getAdapter(String id) {
    SimValueAdapter result = (SimValueAdapter)_adapters.get(id);
    if (result != null) return result;
    result = new SimValueAdapter();
    result.attach(this, id);
    _adapters.put(id, result);
    AttachCommand command = new AttachCommand();
    command.id = id;
    _port.send(command);
    return result;
  }

  public synchronized
  void registerAdapter(String id, SimValueAdapter adapter) {
    _adapters.put(id, adapter);
    AttachCommand command = new AttachCommand();
    command.id = id;
    _port.send(command);
  }

  public synchronized void removeAdapter(String id) {
    SimValueAdapter adapter = (SimValueAdapter)_adapters.remove(id);
    DetachCommand command = new DetachCommand();
    command.id = id;
    _port.send(command);
  }

  public synchronized void accept(ClientPort receiver, ClientCommand command) {
    command.execute(this, _port);
  }

  public void acceptChangeList(Vector changeList) {
    for (Enumeration e = changeList.elements(); e.hasMoreElements(); ) {
      Change change = (Change)e.nextElement();
      getAdapter(change.id).acceptChange(change);
    }
  }

  public void acceptObjectState(String id, Vector stateList) {
    SimValueAdapter adapter = getAdapter(id);
    adapter.resetState();
    for (Enumeration e = stateList.elements(); e.hasMoreElements(); ) {
      adapter.acceptChange((Change)e.nextElement());
    }
  }

  public void getAgentState(String id) {
    GetAgentStateCommand command = new GetAgentStateCommand();
    command.id = id;
    _port.send(command);
  }

  /**
   * ѥ᡼ꥹȤ롥
   * 
   * @param id		֥ȣɣ
   * @param actionName	̾٥ǤϤʤȤա
   * @return		б륢¸ߤġvisibleäȤ
   *	ϥѥ᡼ꥹȤ֤<code>UIObject</code>
   *	󥹥󥹡. ǤʤСnull֤롥
   */
  public CompositeUIObject getParameterList(String id, String actionName)
       throws IOException
  {
    RequestParametersCommand command = new RequestParametersCommand();
    command.id = id;
    command.actionName = actionName;
    CompositeUIObject result =
      (CompositeUIObject)new Sync(_port).call(command);
    if (result == null) {
      throw new IOException("cant get parameters of " + actionName);
    }
    return result;
  }

  /**
   * ư.
   * 
   * @param id			֥ȣɣ
   * @param actionName		̾ʥ٥ǤϤʤȤա
   * @param parameterList	ѥ᡼ꥹ
   * @return	б륢¸ߤġvisibleäȤ
   *	true֤. ǤʤСfalse֤롥
   */
  public boolean invokeAction(String id,
			      String actionName, UIObject parameters)
       throws IOException
  {
    InvokeActionCommand command = new InvokeActionCommand();
    command.id = id;
    command.actionName = actionName;
    command.parameters = parameters;
    Sync sync = new Sync(_port);
    return ((Boolean)sync.call(command)).booleanValue();
  }

  public boolean replyToInteractionRequest(int key, UIObject parameter)
       throws IOException
  {
    ReplyInteractionCommand command = new ReplyInteractionCommand();
    command.key = key;
    command.parameter = parameter;
    return ((Boolean)new Sync(_port).call(command)).booleanValue();
  }
}
