package ORG.ebiss.sara.server;

import java.util.*;
import java.io.*;

public class OldObjectInputStream extends FilterInputStream
implements DataInput {
  private Hashtable		_label2object = new Hashtable();
  private DataInputStream	_rawIn;
  private DataInputStream	_in;
  private boolean		_recurring;

  public OldObjectInputStream(InputStream in) {
    super(in);
    _rawIn = new DataInputStream(in);
    _in = _rawIn;
    _recurring = false;
  }

  public synchronized int read() throws IOException {
    return _in.read();
  }

  public synchronized int read(byte[] b) throws IOException {
    return _in.read(b);
  }

  public synchronized int read(byte[] b, int off, int len) throws IOException {
    return _in.read(b, off, len);
  }

  public synchronized long skip(long n) throws IOException {
    return _in.skip(n);
  }

  public synchronized int available() throws IOException {
    return _in.available();
  }

  public synchronized void close() throws IOException {
    _in.close();
  }

  public synchronized void mark(int readlimit) {
    _in.mark(readlimit);
  }

  public synchronized void reset() throws IOException {
    _in.reset();
  }

  public synchronized boolean markSupported() {
    return _in.markSupported();
  }

  public synchronized void readFully(byte[] b) throws IOException {
    _in.readFully(b);
  }

  public synchronized void readFully(byte[] b, int off, int len)
       throws IOException {
    _in.readFully(b, off, len);
  }

  public synchronized int skipBytes(int n) throws IOException {
    return _in.skipBytes(n);
  }

  public synchronized boolean readBoolean() throws IOException {
    return _in.readBoolean();
  }

  public synchronized byte readByte() throws IOException {
    return _in.readByte();
  }

  public synchronized int readUnsignedByte() throws IOException {
    return _in.readUnsignedByte();
  }

  public synchronized short readShort() throws IOException {
    return _in.readShort();
  }

  public synchronized int readUnsignedShort() throws IOException {
    return _in.readUnsignedShort();
  }

  public synchronized char readChar() throws IOException {
    return _in.readChar();
  }

  public synchronized int readInt() throws IOException {
    return _in.readInt();
  }

  public synchronized long readLong() throws IOException {
    return _in.readLong();
  }

  public synchronized float readFloat() throws IOException {
    return _in.readFloat();
  }

  public synchronized double readDouble() throws IOException {
    return _in.readDouble();
  }

  public synchronized String readLine() throws IOException {
    return _in.readLine();
  }

  public synchronized String readUTF() throws IOException {
    int length = _in.readInt();
    StringBuffer result = new StringBuffer(length);
    for (int i = length; i > 0; --i) {
      result.append(_in.readChar());
    }
    return result.toString();
  }

  public synchronized Object readObject()
       throws IOException, ClassNotFoundException,
	      InstantiationException, IllegalAccessException {
    if (_recurring) {
      int label = readInt();
      Object result = _label2object.get(new Integer(label));
      if (result instanceof PartiallyLoadedObject) {
	return ((PartiallyLoadedObject)result).object;
      }
      return result;
    }

    Object root = null;
    Vector transportables = new Vector();

    Object object, label;
    for (;;) {
      String className = readUTF();
      if (className.length() == 0) break;
      label = new Integer(readInt());
      if (className.equals("java.lang.String")) {
	object = readUTF();
      } else if (className.equals("java.lang.Integer")) {
	object = new Integer(readInt());
      } else if (className.equals("java.lang.Long")) {
	object = new Long(readLong());
      } else if (className.equals("java.lang.Boolean")) {
	object = new Boolean(readBoolean());
      } else if (className.equals("java.lang.Float")) {
	object = new Float(readFloat());
      } else if (className.equals("java.lang.Double")) {
	object = new Double(readDouble());
      } else {
	byte data[] = new byte[readInt()];
	readFully(data);
	try {
	  Transportable body =
	    (Transportable)Class.forName(className).newInstance();
	  object = new PartiallyLoadedObject(body, data);
	  transportables.addElement(object);
	} catch (Throwable e) {
	  e.printStackTrace();
	  throw new RuntimeException();
	}
      }
      if (root == null) root = object;
      _label2object.put(label, object);
    }

    _recurring = true;
    for (Enumeration e = transportables.elements(); e.hasMoreElements(); ) {
      PartiallyLoadedObject v = (PartiallyLoadedObject)e.nextElement();
      _in = new DataInputStream(new ByteArrayInputStream(v.data));
      v.object.readObject(this);
    }
    _in = _rawIn;
    _recurring = false;

    _label2object.clear();
    return root instanceof PartiallyLoadedObject
      ? ((PartiallyLoadedObject)root).object
      : root;
  }
}

class PartiallyLoadedObject {

  public Transportable	object;
  public byte		data[];

  PartiallyLoadedObject(Transportable o, byte d[]) {
    object = o;
    data = d;
  }
}
