package ORG.ebiss.sara.server;

import java.io.*;
import java.util.*;
import COM.objectspace.jgl.*;

public class OldObjectOutputStream extends FilterOutputStream
implements DataOutput {
  private Queue 		_todo;
  private Hashtable 		_object2label;
  private Hashtable 		_label2object;
  private DataOutputStream	_rawOut;
  private DataOutputStream	_bufferedOut;
  private ByteArrayOutputStream	_buffer;
  private DataOutputStream	_out;
  private boolean		_recurring;
  private int			_labelCounter;

  public OldObjectOutputStream(OutputStream out) {
    super(out);
    _todo = new Queue();
    _object2label = new Hashtable();
    _label2object = new Hashtable();
    _buffer = new ByteArrayOutputStream();
    _bufferedOut = new DataOutputStream(_buffer);
    _rawOut = new DataOutputStream(out);
    _out = _rawOut;
    _recurring = false;
    _labelCounter = 0;
  }

  public synchronized void close() throws IOException {
    _out.close();
  }

  public synchronized void flush() throws IOException {
    _out.flush();
  }

  public synchronized void write(int b) throws IOException {
    _out.write(b);
  }

  public synchronized void write(byte[] b) throws IOException {
    _out.write(b);
  }

  public synchronized void write(byte[] b, int off, int len)
       throws IOException {
    _out.write(b, off, len);
  }

  public synchronized void writeBoolean(boolean v) throws IOException {
    _out.writeBoolean(v);
  }

  public synchronized void writeByte(int v) throws IOException {
    _out.writeByte(v);
  }

  public synchronized void writeShort(int v) throws IOException {
    _out.writeShort(v);
  }

  public synchronized void writeChar(int v) throws IOException {
    _out.writeChar(v);
  }

  public synchronized void writeInt(int v) throws IOException {
    _out.writeInt(v);
  }

  public synchronized void writeLong(long v) throws IOException {
    _out.writeLong(v);
  }

  public synchronized void writeFloat(float v) throws IOException {
    _out.writeFloat(v);
  }

  public synchronized void writeDouble(double v) throws IOException {
    _out.writeDouble(v);
  }

  public synchronized void writeBytes(String v) throws IOException {
    _out.writeBytes(v);
  }

  public synchronized void writeChars(String v) throws IOException {
    _out.writeChars(v);
  }

  public synchronized void writeUTF(String v) throws IOException {
    _out.writeInt(v.length());
    _out.writeChars(v);
  }

  public synchronized void writeObject(Object v) throws IOException {
    if (_recurring) {
      Integer label = (Integer)_object2label.get(v);
      if (label == null) {
	label = new Integer(_labelCounter++);
	_object2label.put(v, label);
	Object dup = _label2object.put(label, v);
	if (dup != null) {
	  throw new
	    RuntimeException("the same label is given to many objects");
	}
	_todo.push(v);
      }
      writeInt(label.intValue());
      return;
    }

    Integer label = new Integer(_labelCounter++);
    _object2label.clear();
    _label2object.clear();
    _object2label.put(v, label);
    _label2object.put(label, v);
    _todo.push(v);
    while (!_todo.isEmpty()) {
      v = _todo.pop();
      writeUTF(v.getClass().getName());
      writeInt(((Integer)_object2label.get(v)).intValue());
      if (v instanceof String) {
	writeUTF((String)v);
      } else if (v instanceof Integer) {
	writeInt(((Integer)v).intValue());
      } else if (v instanceof Long) {
	writeLong(((Long)v).longValue());
      } else if (v instanceof Boolean) {
	writeBoolean(((Boolean)v).booleanValue());
      } else if (v instanceof Float) {
	writeFloat(((Float)v).floatValue());
      } else if (v instanceof Double) {
	writeDouble(((Double)v).doubleValue());
      } else {
	_recurring = true;
	_out = _bufferedOut;
	((Transportable)v).writeObject(this);
	_out.flush();
	_recurring = false;
	_out = _rawOut;
	int size = _buffer.size();
	write(0xff & (size >> 24));
	write(0xff & (size >> 16));
	write(0xff & (size >> 8));
	write(0xff & size);
	_buffer.writeTo(_rawOut);
	_buffer.reset();
      }
    }
    writeUTF("");
    _object2label.clear();
    _label2object.clear();
  }
}
