package ORG.ebiss.sara.server;

import java.io.*;
import java.net.*;
import java.util.*;
import COM.objectspace.jgl.*;
import ORG.ebiss.sara.interpreter.*;

/**
 * ̿ݡ
 * 
 * @see	
 * @version	1.0 Oct 1997
 * @author	TOGAWA Atsushi
 */
public class ServerPort extends Port {
  protected Server	_server;
  protected HashSet	_subjects = new HashSet();

  public ServerPort(Server server, Socket socket) {
    _server = server;
    open(socket);
  }

  protected Receiver createReceiver() throws IOException {
    return new ServerReceiver(this, _socket);
  }

  public void accept(Command command) {
    _server.accept(this, (ServerCommand)command);
  }

  public void addSubject(String id) { _subjects.add(id); }

  public void removeSubject(String id) { _subjects.remove(id); }

  public Enumeration subjects() { return _subjects.elements(); }

  public class ServerReceiver extends Receiver {
    ServerReceiver(Port port, Socket socket) throws IOException {
      super(port, socket);
    }

    public void run() {
      try {
	_server.registerPort(ServerPort.this);
	super.run();
      } finally {
	_server.unregisterPort(ServerPort.this);
      }
    }
  }

  public void send(String id, ClientCommand command) {
    if (_subjects.get(id) != null) send(command);
  }

  public void sendChangeList(Container changeList) {
    if (changeList.size() <= 0) return;
    AcceptChangeListCommand command = new AcceptChangeListCommand();
    for (Enumeration e = changeList.elements(); e.hasMoreElements(); ) {
      Value value = (Value)e.nextElement();
      try {
	if (_subjects.get(value.read("id").stringValue()) != null) {
	  command.changeList.addElement(Server.createChangeRecord(value));
	}
      } catch (CoerceFailedException cfe) {
	cfe.printStackTrace();
	throw new RuntimeException();
      }
    }
    send(command);
  }

  public void attachRemoteObserver(Interpreter interpreter, final String id) {
    addSubject(id);
    try {
      Value postOffice = interpreter.readVariable("thePostOffice");
      ApplicableValue passStateList =
	(ApplicableValue) postOffice.read("passStateList");
      Value[] arguments = new Value[2];
      arguments[0] = SymbolValue.intern(id);
      arguments[1] = new ApplicableValue() {
	public Code apply(Interpreter interpreter, int[] arguments,
			  int resultSlot, Code returnPoint) {
	  ContainerValue result =
	    (ContainerValue)interpreter.readRegister(arguments[0]);
	  Container container = (Container)result.getJavaValue();
	  if (container.size() <= 0) return returnPoint;

	  AcceptChangeListCommand command = new AcceptChangeListCommand();
	  for (Enumeration e = container.elements(); e.hasMoreElements(); ) {
	    Value value = (Value)e.nextElement();
	    command.changeList.addElement(Server.createChangeRecord(value));
	  }
	  send(command);
	  return returnPoint;
	}

	public boolean checkArity(int arity) { return arity == 1; }
      };
      interpreter.invoke(passStateList, arguments);
    } catch (EvaluationException ee) {
      ee.print(interpreter, _server.out);
    }
  }

  public void show(String id, UIObject contents) {
    if (_subjects.get(id) == null) return;
    ShowCommand command = new ShowCommand();
    command.contents = contents;
    command.id = id;
    send(command);
  }

  public static class ShowCommand extends ClientCommand {
    public UIObject	contents;
    public String		id;

    public void writeObject(OldObjectOutputStream out) throws IOException {
      super.writeObject(out);
      out.writeObject(contents);
      out.writeUTF(id);
    }

    public void readObject(OldObjectInputStream in)
	 throws IOException, ClassNotFoundException, IllegalAccessException,
	   InstantiationException
    {
      super.readObject(in);
      contents = (UIObject)in.readObject();
      id = in.readUTF();
    }

    public void execute() {
      ((SimValueAdapter)client.getAdapter(id)).show(contents);
    }
  }

  public void interact(String id,
		       ClientCommand command, Continuation whenReplied)
  {
    if (_subjects.get(id) == null) return;
    request(command, whenReplied);
  }
}
