package ORG.ebiss.sara.server;

import java.io.*;
import java.net.*;
import java.util.*;
import COM.objectspace.jgl.*;

public class SimValueAdapter {
  protected Client	_client;
  protected String	_objectId;
  protected Vector	_observers = new Vector(2);
  protected Hashtable	_visible = new Hashtable(37);
  protected Hashtable	_actionInteractions = new Hashtable(19);
  protected Hashtable	_queryInteractions = new Hashtable(19);
  protected Vector	_info = new Vector();

  protected void attach(Client client, String objectId) {
    _client = client;
    _objectId = objectId;
  }

  public synchronized void addObserver(RemoteObserver observer) {
    _observers.addElement(observer);
    for (Enumeration e = _visible.elements(); e.hasMoreElements(); ) {
      Change change = (Change)e.nextElement();
      makeVisible(change);
    }
    for (Enumeration e = _info.elements(); e.hasMoreElements(); ) {
      Pair pair = (Pair)e.nextElement();
      observer.acceptExtendedInformation(this, (String[])pair.first,
					 (UIObject)pair.second);
    }
  }

  public synchronized void removeObserver(RemoteObserver observer) {
    _observers.removeElement(observer);
  }

  public synchronized void resetState() {
    for (Enumeration e = _visible.elements(); e.hasMoreElements(); ) {
      makeInvisible((Change)e.nextElement());
    }
    _visible.clear();
  }

  public synchronized void acceptChange(Change change) {
    switch (change.type) {
    case Change.ADD:
      if (change.visible) {
	if (_visible.put(change.slotName, change) == null) makeVisible(change);
      } else {
	if (_visible.remove(change.slotName) != null) makeInvisible(change);
      }
      break;

    case Change.MODIFY:
    case Change.DUMP:
      if (change.visible) {
	_visible.put(change.slotName, change);
	makeVisible(change);
      } else {
	if (_visible.remove(change.slotName) != null) makeInvisible(change);
      }
      break;

    case Change.REMOVE:
      if (_visible.remove(change.slotName) != null) makeInvisible(change);
      break;

    default:
      throw new RuntimeException(Integer.toString(change.type));
    }
  }

  protected void makeVisible(Change change) {
    if (change.action) {
      ActionInteraction interaction =
	(ActionInteraction)_actionInteractions.get(change.slotName);
      if (interaction == null) {
	interaction = new ActionInteraction(_client, _objectId,
					    change.label, change.slotName);
	_actionInteractions.put(change.slotName, interaction);
      }
      for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
	((RemoteObserver)e.nextElement()).acceptInteraction(this, interaction);
      }
    } else {
      for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
	((RemoteObserver)e.nextElement()).
	  acceptProperty(this, change.slotName, change.value);
      }
    }
  }

  protected void makeInvisible(Change change) {
    if (change.action) {
      ActionInteraction interaction =
	(ActionInteraction)_actionInteractions.remove(change.slotName);
      if (interaction == null) return;
      for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
	((RemoteObserver)e.nextElement()).
	  retractInteraction(this, interaction);
      }
    } else {
      for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
	System.out.println("retract: " + change.slotName + "@" + _objectId);
	((RemoteObserver)e.nextElement()).
	  retractProperty(this, change.slotName);
      }
    }
  }

  public boolean execute(PlayersAction action) {
    try {
      return action.execute(_client);
    } catch (IOException ioe) {
      return false;
    }
  }

  public void show(UIObject contents) {
    for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
      ((RemoteObserver)e.nextElement()).acceptInformation(this, contents);
    }
  }

  public synchronized void acceptInteractionRequest(int key,
	CompositeUIObject parameter)
  {
    Integer keyObj = new Integer(key);
    if (_queryInteractions.get(keyObj) != null) return;

    QueryInteraction interaction = new QueryInteraction(key, parameter);
    interaction.key = key;
    interaction.parameter = parameter;
    _queryInteractions.put(keyObj, interaction);
    for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
      ((RemoteObserver)e.nextElement()).acceptInteraction(this, interaction);
    }
  }

  public synchronized void retractInteractionRequest(int key) {
    QueryInteraction interaction =
      (QueryInteraction)_queryInteractions.get(new Integer(key));
    if (interaction == null) return;
    for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
      ((RemoteObserver)e.nextElement()).retractInteraction(this, interaction);
    }
  }

  public synchronized
  void acceptExtendedInformation(String[] type, UIObject info) {
    Pair pair = new Pair(type, info);
    _info.addElement(pair);
    for (Enumeration e = _observers.elements(); e.hasMoreElements(); ) {
      ((RemoteObserver)e.nextElement()).
	acceptExtendedInformation(this, type, info);
    }
  }
}
