package ORG.ebiss.sara.server;

import java.io.*;
import java.util.*;
import ORG.ebiss.sara.interpreter.*;

public class UICheckBox implements CompositeUIObject {
  public String	description;
  public Vector	labels;
  public Vector	checks;

  public void writeObject(OldObjectOutputStream out) throws IOException {
    out.writeUTF(description);
    out.writeInt(labels.size());
    for (Enumeration e = labels.elements(); e.hasMoreElements(); ) {
      out.writeObject(e.nextElement());
    }
    out.writeInt(checks.size());
    for (Enumeration e = checks.elements(); e.hasMoreElements(); ) {
      out.writeObject(e.nextElement());
    }
  }

  public void readObject(OldObjectInputStream in)
       throws IOException, ClassNotFoundException, IllegalAccessException,
	 InstantiationException
  {
    description = in.readUTF();
    int n = in.readInt();
    labels = new Vector(n);
    for (int i = 0; i < n; ++i) {
      labels.addElement(in.readObject());
    }
    n = in.readInt();
    checks = new Vector(n);
    for (int i = 0; i < n; ++i) {
      checks.addElement(in.readObject());
    }
  }

  public String getDescription() { return description; }

  public int getType() { return IS_A_CHECKBOX; }

  public void setValueOf(String value) {
    throw new RuntimeException("unsupported operation: setValueOf(String)");
  }

  public String toString() {
    StringBuffer buffer = new StringBuffer("[");
    int n = labels.size();
    for (int i = 0; i < n; ++i) {
      if (i > 0) buffer.append(", ");
      buffer.append(labels.elementAt(i));
      buffer.append(((Boolean)checks.elementAt(i)).booleanValue()
		    ? ": true" : ": false");
    }
    buffer.append("]");
    return buffer.toString();
  }

  public void check(String label, boolean f) {
    int n = labels.size();
    for (int i = 0; i < n; ++i) {
      if (labels.elementAt(i).toString().equals(label)) {
	((Vector)checks).setElementAt(f ? Boolean.TRUE : Boolean.FALSE, i);
      }
    }
    // throw new NoSuchElementException();
  }

  public boolean isChecked(String label) {
    int n = labels.size();
    for (int i = 0; i < n; ++i) {
      if (labels.elementAt(i).toString().equals(label)) {
	return ((Boolean)((Vector)checks).elementAt(i)).booleanValue();
      }
    }
    // throw new NoSuchElementException();
    return false;
  }

  public Enumeration elements() {
    return new UICheckBoxEnumerator(labels.elements());
  }
}

class UICheckBoxEnumerator implements Enumeration {
  Enumeration	labels;

  UICheckBoxEnumerator(Enumeration labels) {
    this.labels = labels;
  }

  public boolean hasMoreElements() { return labels.hasMoreElements(); }

  public Object nextElement() { return labels.nextElement().toString(); }
}
