package ORG.ebiss.sara.server;

import java.io.*;
import java.util.*;

public class UIChoice implements CompositeUIObject {
  public String	description;
  public int	index;
  public Vector	labels;
  public Vector	values;

  public void writeObject(OldObjectOutputStream out) throws IOException {
    out.writeUTF(description);
    out.writeInt(index);
    out.writeInt(labels.size());
    for (Enumeration e = labels.elements(); e.hasMoreElements(); ) {
      out.writeObject(e.nextElement());
    }
    out.writeInt(values.size());
    for (Enumeration e = values.elements(); e.hasMoreElements(); ) {
      out.writeObject(e.nextElement());
    }
  }

  public void readObject(OldObjectInputStream in)
       throws IOException, ClassNotFoundException, IllegalAccessException,
	 InstantiationException
  {
    description = in.readUTF();
    index = in.readInt();
    int n = in.readInt();
    labels = new Vector(n);
    for (int i = 0; i < n; ++i) {
      labels.addElement(in.readObject());
    }
    n = in.readInt();
    values = new Vector(n);
    for (int i = 0; i < n; ++i) {
      values.addElement(in.readObject());
    }
  }

  public String toString() {
    return labels.size() == 0 ? "" : labels.elementAt(index).toString();
  }

  public String getDescription() { return description; }

  public int getType() { return IS_A_CHOICE; }

  public Enumeration elements() {
    return new UIChoiceEnumerator(labels.elements());
  }

  public void setValueOf(String value) {
    int i = 0;
    for (Enumeration e = labels.elements(); e.hasMoreElements(); ++i) {
      if (e.nextElement().toString().equals(value)) {
	index = i;
	return;
      }
    }
    throw new RuntimeException("invalid value");
  }
}

class UIChoiceEnumerator implements Enumeration {
  Enumeration	labels;

  UIChoiceEnumerator(Enumeration labels) {
    this.labels = labels;
  }

  public boolean hasMoreElements() { return labels.hasMoreElements(); }

  public Object nextElement() { return labels.nextElement().toString(); }
}
